"""
Instrument driver for the Newport SMC100CC motion controller.
"""
import logging
from time import sleep
from typing import Dict, Optional
import enum

from qmi.core.context import QMI_Context
from qmi.core.exceptions import QMI_InstrumentException
from qmi.core.rpc import rpc_method
from qmi.instruments.newport.actuators import LinearActuator
from qmi.instruments.newport.single_axis_motion_controller import Newport_SingleAxisMotionController

# Global variable holding the logger for this module.
_logger = logging.getLogger(__name__)


class ControlLoopState(enum.IntEnum):
    """Control loop states of the controller."""
    OPEN = 0
    CLOSED = 1


class Newport_SMC100CC(Newport_SingleAxisMotionController):
    """Instrument driver for the Newport SMC100CC servo motion controller."""

    def __init__(self,
                 context: QMI_Context,
                 name: str,
                 transport: str,
                 serial: str,
                 actuators: Dict[Optional[int], LinearActuator],
                 baudrate: int = 57600) -> None:
        """Initialize driver.

        Parameters:
            name:       Name for this instrument instance.
            transport:  QMI transport descriptor to connect to the instrument.
            serial:     The serial number of the instrument.
            actuators:  The linear actuators that this controller will drive. Each controller address
                        drives a linear actuator. The key of the dictionary is the controller address
                        and the value is the actuator that it drives.
            baudrate:   The baudrate of the instrument. Defaults to 57600.
        """
        super().__init__(context, name, transport, serial, actuators, baudrate)

    def _get_configuration_or_disable(self, parameter: str) -> bool:
        """Check the current state and try to enter DISABLE state. Failing that, enter CONFIGURATION state.

        Parameters:
            parameter: String indicating which controller command is being used for the check.

        Returns:
            persist: True if the entered state is CONFIGURATION state. False if the entered state is DISABLE state.
        """
        try:
            persist = False
            istate = self._state_ready_check(parameter)
            if istate < int("3C", 16):
                self._enter_disable_state()

        except QMI_InstrumentException:
            self._enter_configuration_state()
            persist = True

        return persist

    def _set_configuration_or_disable(self, parameter: str, persist: bool):
        """The state must be either CONFIGURATION or DISABLE.

        Parameters:
            parameter: String to indicate which command parameter is being checked
            persist: If True, force the CONFIGURATION state. If False, try to enter DISABLE state.
        """
        if persist:
            # instrument must be in CONFIGURATION or DISABLE state to get the maximum allowed following error.
            self._enter_configuration_state()

        else:
            istate = self._state_ready_check(parameter)
            # If the state is READY (32-38), we must enter DISABLE state first
            if istate < int("3C", 16):
                self._enter_disable_state()

    def _exit_configuration_or_disable(self, exit_persist: bool):
        """Exit from either CONFIGURATION or DISABLE state to NOT REFERENCED or READY state, respectively.

        Parameters:
            exit_persist: True to exit from CONFIGURATION state to NOT REFERENCED state.
                               False to exit from DISABLE state to READY state.
        """
        if exit_persist:
            self._exit_configuration_state()

        else:
            self._exit_disable_state()

    @rpc_method
    def get_encoder_increment_value(self, controller_address: Optional[int] = None) -> float:
        """
        Get the encoder increment value.

        Parameters:
            controller_address: Optional address of the controller that needs to be controlled. By default,
                                it is set to the initialised value of the controller address.

        Returns:
            Encoder increment value.
        """
        self.controller_address = controller_address
        _logger.info(
            "Getting encoder increment value of instrument [%s]", self._name)
        self._enter_configuration_state()
        res = self._scpi_protocol.ask(
            self._build_command("SU?"))
        self._check_error()
        self._exit_configuration_state()
        return self._actuators[self.controller_address].ENCODER_RESOLUTION / float(res[3:])

    @rpc_method
    def set_encoder_increment_value(self, value: float, controller_address: Optional[int] = None) -> None:
        """
        Set the encoder increment value. By default, to be as close to 1mm as possible.
        Check the example (SU command) in the doc below to see how the increment value is calculated:
        https://www.newport.com/mam/celum/celum_assets/np/resources/CONEX-CC_-_Controller_Documentation.pdf?0

        Parameters:
            value:              Increment value.
            controller_address: Optional address of the controller that needs to be controlled. By default,
                                it is set to the initialised value of the controller address.
        """
        self.controller_address = controller_address
        _logger.info(
            "Setting encoder increment value of instrument [%s] to [%f]", self._name, value)
        self._enter_configuration_state()
        self._scpi_protocol.write(self._build_command(
            "SU", value))
        sleep(self.COMMAND_EXEC_TIME)
        self._check_error()
        self._exit_configuration_state()

    @rpc_method
    def get_driver_voltage(self, controller_address: Optional[int] = None) -> float:
        """
        Get the max. output voltage of the driver to the motor.

        Parameters:
            controller_address: Optional address of the controller that needs to be controlled. By default,
                                it is set to the initialised value of the controller address.
        """
        self.controller_address = controller_address
        _logger.info(
            "Getting the max. output voltage of the driver to the motor of instrument [%s]", self._name)
        self._enter_configuration_state()
        driver_voltage = self._scpi_protocol.ask(
            self._build_command("DV?"))
        self._check_error()
        self._exit_configuration_state()
        return float(driver_voltage[3:])

    @rpc_method
    def set_driver_voltage(self, driver_voltage: float, controller_address: Optional[int] = None) -> None:
        """
        Set the max. output voltage of the driver to the motor.

        Parameters:
            driver_voltage:     New motor driver voltage in Volts. Must be 12 V <= driver_voltage <= 48 V.
            controller_address: Optional address of the controller that needs to be controlled. By default,
                                it is set to the initialised value of the controller address.
        """
        if driver_voltage < 12.0 or driver_voltage > 48.0:
            raise QMI_InstrumentException(
                f"Provided value {driver_voltage} not in valid range 12 <= driver_voltage <= 48.")

        self.controller_address = controller_address
        _logger.info(
            "Setting the max. output voltage of the driver to the motor of instrument [%s] to [%f]",
            self._name, driver_voltage
        )
        self._enter_configuration_state()
        self._scpi_protocol.write(self._build_command(
            "DV", driver_voltage))
        sleep(self.COMMAND_EXEC_TIME)
        self._check_error()
        self._exit_configuration_state()

    @rpc_method
    def get_low_pass_filter_cutoff_frequency(self, controller_address: Optional[int] = None) -> float:
        """
        Get the low pass filter cut-off frequency Kd.

        Parameters:
            controller_address: Optional address of the controller that needs to be controlled. By default,
                                it is set to the initialised value of the controller address.

        Returns:
            The low pass filter cut-off frequency, Kd, in Hertz.
        """
        self.controller_address = controller_address
        _logger.info(
            "Getting encoder increment value of instrument [%s]", self._name)
        persist = self._get_configuration_or_disable("cut-off frequency")
        res = self._scpi_protocol.ask(
            self._build_command("FD?"))
        self._check_error()
        self._exit_configuration_or_disable(persist)

        return float(res[3:])

    @rpc_method
    def set_low_pass_filter_cutoff_frequency(
            self, frequency: float, persist: bool = False, controller_address: Optional[int] = None) -> None:
        """
        Set the low pass filter cut-off frequency Kd. Can be set only in CONFIGURATION or DISABLE states.

        Parameters:
            frequency:          Cutoff frequency Kd in Hertz. Must be 1E-6 < frequency < 2000.
            persist:            Flag to indicate if the frequency cutoff should be persisted to the controller's memory,
                                so it is still available after powering down the controller. When not persisted, the
                                frequency cutoff is the one stored in the controller's memory.
            controller_address: Optional address of the controller that needs to be controlled. By default,
                                it is set to the initialised frequency of the controller address.
        """
        if frequency <= self.MIN_FLOAT_LIMIT or frequency >= 2000:
            raise QMI_InstrumentException(
                f"Provided value {frequency} not in valid range {self.MIN_FLOAT_LIMIT} < frequency < 2000.")

        self.controller_address = controller_address
        _logger.info(
            "Setting encoder increment frequency of instrument [%s] to [%s]", self._name, frequency)
        self._set_configuration_or_disable("cut-off frequency", persist)
        self._scpi_protocol.write(self._build_command(
            "FD", frequency))
        sleep(self.COMMAND_EXEC_TIME)
        self._check_error()
        self._exit_configuration_or_disable(persist)

    @rpc_method
    def get_following_error_limit(self, controller_address: Optional[int] = None) -> float:
        """
        Get the maximum allowed following error.

        Parameters:
            controller_address: Optional address of the controller that needs to be controlled. By default,
                                it is set to the initialised value of the controller address.

        Returns:
            The value for the maximum allowed following error.
        """
        self.controller_address = controller_address
        _logger.info(
            "Getting maximum allowed following error of instrument [%s]", self._name)
        persist = self._get_configuration_or_disable("maximum allowed following error")
        res = self._scpi_protocol.ask(
            self._build_command("FE?"))
        sleep(self.COMMAND_EXEC_TIME)
        self._check_error()
        self._exit_configuration_or_disable(persist)

        return float(res[3:])

    @rpc_method
    def set_following_error_limit(
            self, error_limit: float, persist: bool = False, controller_address: Optional[int] = None) -> None:
        """
        Set the value for the maximum allowed following error. It can be set only in CONFIGURATION or DISABLE state.

        Parameters:
            error_limit:        The value for the maximum allowed following error. Must be 1E-6 < error_limit < 1E12.
            persist:            Flag to indicate if the error limit should be persisted to the controller's memory, so
                                it is still available after powering down the controller. When not persisted, the error
                                limit is the one stored in the controller's memory.
            controller_address: Optional address of the controller that needs to be controlled. By default,
                                it is set to the initialised error limit of the controller address.
        """
        if error_limit <= self.MIN_FLOAT_LIMIT or error_limit >= self.MAX_FLOAT_LIMIT:
            raise QMI_InstrumentException(
                f"Provided value {error_limit} not in valid range {self.MIN_FLOAT_LIMIT} < error limit "
                f"< {self.MAX_FLOAT_LIMIT}."
            )

        self.controller_address = controller_address
        _logger.info(
            "Setting the value for the maximum allowed following error of instrument [%s] to [%f]",
            self._name, error_limit
        )
        self._set_configuration_or_disable("following error limit", persist)
        self._scpi_protocol.write(self._build_command(
            "FE", error_limit))
        sleep(self.COMMAND_EXEC_TIME)
        self._check_error()
        self._exit_configuration_or_disable(persist)

    @rpc_method
    def get_friction_compensation(self, controller_address: Optional[int] = None) -> float:
        """
        Get the friction compensation.

        Parameters:
            controller_address: Optional address of the controller that needs to be controlled. By default,
                                it is set to the initialised value of the controller address.
        """
        self.controller_address = controller_address
        _logger.info(
            "Getting the friction compensation of instrument [%s]", self._name)
        persist = self._get_configuration_or_disable("friction compensation")
        friction_compensation = self._scpi_protocol.ask(
            self._build_command("FF?"))
        self._check_error()
        self._exit_configuration_or_disable(persist)

        return float(friction_compensation[3:])

    @rpc_method
    def set_friction_compensation(
            self, friction_compensation: float, persist: bool = False, controller_address: Optional[int] = None
    ) -> None:
        """
        Set the friction compensation. It must not be larger than the driver voltage set by DV command. We can check
        the input value for correct range in CONFIGURATION state, but in READY/DISABLE state we cannot as CONFIGURATION
        state is needed to check the driver voltage. Exiting the CONFIGURATION state leads to NOT REFERENCED state
        which cannot then be changed to READY or DISABLE state.

        Parameters:
            friction_compensation: New friction compensation value.
                                   Must be 0 <= friction_compensation < driver_voltage.
            persist:               Flag to indicate if the friction compensation should be persisted to the controller's
                                   memory, so it is still available after powering down the controller. When not
                                   persisted, the friction compensation is the one stored in the controller's memory.
            controller_address:    Optional address of the controller that needs to be controlled. By default,
                                   it is set to the initialised value of the controller address.
        """
        self.controller_address = controller_address
        if persist:
            # First we can check the driver voltage to get valid range for friction compensation.
            self._enter_configuration_state()
            response = self._scpi_protocol.ask(
                self._build_command("DV?"))
            self._check_error()
            driver_voltage = float(response[3:])
            if friction_compensation < 0 or friction_compensation >= driver_voltage:
                self._exit_configuration_state()
                raise QMI_InstrumentException(
                    f"Provided value {friction_compensation} not in valid range 0 <= "
                    f"friction_compensation < {driver_voltage}."
                )

        _logger.info(
            "Setting the friction compensation of instrument [%s] to [%f]", self._name, friction_compensation)
        if not persist:
            # instrument must be in DISABLE state to set the friction compensation.
            istate = self._state_ready_check("friction compensation")
            # If the state is READY (32-38), we must enter DISABLE state first
            if istate < int("3C", 16):
                self._enter_disable_state()

        self._scpi_protocol.write(self._build_command(
            "FF", friction_compensation))
        sleep(self.COMMAND_EXEC_TIME)
        self._check_error()
        self._exit_configuration_or_disable(persist)

    @rpc_method
    def get_derivative_gain(self, controller_address: Optional[int] = None) -> float:
        """
        Get the derivative gain of the PID control loop.

        Parameters:
            controller_address: Optional address of the controller that needs to be controlled. By default,
                                it is set to the initialised value of the controller address.

        Returns:
            The derivative gain of the PID control loop.
        """
        self.controller_address = controller_address
        _logger.info(
            "Getting derivative gain of the PID control loop of instrument [%s]", self._name)
        persist = self._get_configuration_or_disable("derivative gain")
        res = self._scpi_protocol.ask(
            self._build_command("KD?"))
        self._check_error()
        self._exit_configuration_or_disable(persist)

        return float(res[3:])

    @rpc_method
    def set_derivative_gain(
            self, derivative_gain: float, persist: bool = False, controller_address: Optional[int] = None) -> None:
        """
        Set the derivative gain of the PID control loop.

        Parameters:
            derivative_gain:    New derivative_gain value in Volt * second/preset unit.
            persist:            Flag to indicate if the derivative gain should be persisted to the controller's memory,
                                so it is still available after powering down the controller. When not persisted, the
                                derivative gain is the one stored in the controller's memory.
            controller_address: Optional address of the controller that needs to be controlled. By default,
                                it is set to the initialised derivative_gain of the controller address.
        """
        if derivative_gain < 0 or derivative_gain >= self.MAX_FLOAT_LIMIT:
            raise QMI_InstrumentException(
                f"Provided value {derivative_gain} not in valid range 0 <= derivative_gain < {self.MAX_FLOAT_LIMIT}.")

        self.controller_address = controller_address
        _logger.info(
            "Setting derivative gain of the PID control loop of instrument [%s] to [%f]",
            self._name, derivative_gain
        )
        self._set_configuration_or_disable("derivative gain", persist)
        self._scpi_protocol.write(self._build_command(
            "KD", derivative_gain))
        sleep(self.COMMAND_EXEC_TIME)
        self._check_error()
        self._exit_configuration_or_disable(persist)

    @rpc_method
    def get_integral_gain(self, controller_address: Optional[int] = None) -> float:
        """
        Get the integral gain of the PID control loop.

        Parameters:
            controller_address: Optional address of the controller that needs to be controlled. By default,
                                it is set to the initialised value of the controller address.

        Returns:
            The integral gain of the PID control loop.
        """
        self.controller_address = controller_address
        _logger.info(
            "Getting integral gain of the PID control loop of instrument [%s]", self._name)
        persist = self._get_configuration_or_disable("integral gain")
        res = self._scpi_protocol.ask(
            self._build_command("KI?"))
        self._check_error()
        self._exit_configuration_or_disable(persist)

        return float(res[3:])

    @rpc_method
    def set_integral_gain(
            self, integral_gain: float, persist: bool = False, controller_address: Optional[int] = None) -> None:
        """
        Set the integral gain of the PID control loop.

        Parameters:
            integral_gain:    New integral_gain value in Volt * second/preset unit.
            persist:            Flag to indicate if the integral gain should be persisted to the controller's memory,
                                so it is still available after powering down the controller. When not persisted, the
                                integral gain is the one stored in the controller's memory.
            controller_address: Optional address of the controller that needs to be controlled. By default,
                                it is set to the initialised integral_gain of the controller address.
        """
        if integral_gain < 0 or integral_gain >= self.MAX_FLOAT_LIMIT:
            raise QMI_InstrumentException(
                f"Provided value {integral_gain} not in valid range 0 <= integral_gain < {self.MAX_FLOAT_LIMIT}.")

        self.controller_address = controller_address
        _logger.info(
            "Setting integral gain of the PID control loop of instrument [%s] to [%f]",
            self._name, integral_gain
        )
        self._set_configuration_or_disable("integral gain", persist)
        self._scpi_protocol.write(self._build_command(
            "KI", integral_gain))
        sleep(self.COMMAND_EXEC_TIME)
        self._check_error()
        self._exit_configuration_or_disable(persist)

    @rpc_method
    def get_proportional_gain(self, controller_address: Optional[int] = None) -> float:
        """
        Get the proportional gain of the PID control loop.

        Parameters:
            controller_address: Optional address of the controller that needs to be controlled. By default,
                                it is set to the initialised value of the controller address.

        Returns:
            The proportional gain of the PID control loop.
        """
        self.controller_address = controller_address
        _logger.info(
            "Getting proportional gain of the PID control loop of instrument [%s]", self._name)
        persist = self._get_configuration_or_disable("proportional gain")
        res = self._scpi_protocol.ask(
            self._build_command("KP?"))
        self._check_error()
        self._exit_configuration_or_disable(persist)

        return float(res[3:])

    @rpc_method
    def set_proportional_gain(
            self, proportional_gain: float, persist: bool = False, controller_address: Optional[int] = None) -> None:
        """
        Set the proportional gain of the PID control loop.

        Parameters:
            proportional_gain:  New proportional_gain value in Volt * second/preset unit.
            persist:            Flag to indicate if the proportional gain should be persisted to the controller's
                                memory, so it is still available after powering down the controller. When not persisted,
                                the proportional gain is the one stored in the controller's memory.
            controller_address: Optional address of the controller that needs to be controlled. By default,
                                it is set to the initialised proportional_gain of the controller address.
        """
        if proportional_gain < 0 or proportional_gain >= self.MAX_FLOAT_LIMIT:
            raise QMI_InstrumentException(
                f"Provided value {proportional_gain} not in valid range 0 <= proportional_gain "
                f"< {self.MAX_FLOAT_LIMIT}."
            )

        self.controller_address = controller_address
        _logger.info(
            "Setting proportional gain of the PID control loop of instrument [%s] to [%f]",
            self._name, proportional_gain
        )
        self._set_configuration_or_disable("proportional gain", persist)
        self._scpi_protocol.write(self._build_command(
            "KP", proportional_gain))
        sleep(self.COMMAND_EXEC_TIME)
        self._check_error()
        self._exit_configuration_or_disable(persist)

    @rpc_method
    def get_velocity_feed_forward(self, controller_address: Optional[int] = None) -> float:
        """
        Get the velocity feed forward of the PID control loop.

        Parameters:
            controller_address: Optional address of the controller that needs to be controlled. By default,
                                it is set to the initialised value of the controller address.

        Returns:
            The velocity feed forward of the PID control loop.
        """
        self.controller_address = controller_address
        _logger.info(
            "Getting velocity feed forward of the PID control loop of instrument [%s]", self._name)
        persist = self._get_configuration_or_disable("velocity feed forward")
        res = self._scpi_protocol.ask(
            self._build_command("KV?"))
        self._check_error()
        self._exit_configuration_or_disable(persist)

        return float(res[3:])

    @rpc_method
    def set_velocity_feed_forward(
            self, velocity_feed_forward: float, persist: bool = False, controller_address: Optional[int] = None
    ) -> None:
        """
        Set the velocity feed forward of the PID control loop.

        Parameters:
            velocity_feed_forward: New velocity_feed_forward value in Volt * second/preset unit.
            persist:            Flag to indicate if the velocity feed forward should be persisted to the controller's
                                memory, so it is still available after powering down the controller. When not persisted,
                                the velocity feed forward is the one stored in the controller's memory.
            controller_address: Optional address of the controller that needs to be controlled. By default,
                                it is set to the initialised velocity_feed_forward of the controller address.
        """
        if velocity_feed_forward < 0 or velocity_feed_forward >= self.MAX_FLOAT_LIMIT:
            raise QMI_InstrumentException(
                f"Provided value {velocity_feed_forward} not in valid range 0 <= velocity_feed_forward "
                f"< {self.MAX_FLOAT_LIMIT}."
            )

        self.controller_address = controller_address
        _logger.info(
            "Setting velocity feed forward of the PID control loop of instrument [%s] to [%f]",
            self._name, velocity_feed_forward
        )
        self._set_configuration_or_disable("velocity feed forward", persist)
        self._scpi_protocol.write(self._build_command(
            "KV", velocity_feed_forward))
        sleep(self.COMMAND_EXEC_TIME)
        self._check_error()
        self._exit_configuration_or_disable(persist)

    @rpc_method
    def get_control_loop_state(self, controller_address: Optional[int] = None) -> ControlLoopState:
        """
        Get the current state of the control loop.

        Parameters:
            controller_address: Optional address of the controller that needs to be controlled. By default,
                                it is set to the initialised value of the controller address.

        Returns:
            control_loop_state: The current control loop state.
        """
        _logger.info(
            "Getting the current state of the control loop of instrument [%s]", self._name)
        self.controller_address = controller_address
        persist = self._get_configuration_or_disable("control loop state")
        control_loop_state = self._scpi_protocol.ask(
            self._build_command("SC?"))
        self._check_error()
        self._exit_configuration_or_disable(persist)

        return ControlLoopState(int(control_loop_state[3:]))

    @rpc_method
    def set_control_loop_state(self, control_loop_state: int, controller_address: Optional[int] = None) -> None:
        """
        Set the current state of the control loop.

        Parameters:
            control_loop_state: New state for the control loop. 0 is OPEN, 1 is CLOSED.
            controller_address: Optional address of the controller that needs to be controlled. By default,
                                it is set to the initialised value of the controller address.
        """
        cls_values = [c.value for c in set(ControlLoopState)]
        if control_loop_state not in cls_values:
            raise QMI_InstrumentException(
                f"Provided value {control_loop_state} not in valid range {cls_values}.")

        self.controller_address = controller_address
        _logger.info(
            "Setting the state of the control loop of instrument [%s] to [%s]",
            self._name, ControlLoopState(control_loop_state).name
        )
        persist = self._get_configuration_or_disable("control loop state")
        self._scpi_protocol.write(self._build_command(
            "SC", control_loop_state))
        sleep(self.COMMAND_EXEC_TIME)
        self._check_error()
        self._exit_configuration_or_disable(persist)



#! /usr/bin/env python3

"""Test DataSet class."""

import io
import unittest

import numpy as np
import h5py

import qmi.data.dataset
from qmi.data.dataset import DataSet


def _internal_create_dataset(name, shape, dtype, add_labels, add_scale, add_attributes):
    """Internal helper function to create a dataset."""

    if np.dtype(dtype).kind == 'i':
        data = dtype(1000 * np.random.random(shape))
    else:
        data = dtype(1 - 2 * np.random.random(shape))

    ds = DataSet(name, data=data)

    if add_labels:
        for axis in range(len(shape) - 1):
            ds.set_axis_label(axis, "axis {} label".format(axis))
            ds.set_axis_unit(axis, "axis {} unit".format(axis))
        for col in range(shape[-1]):
            ds.set_column_label(col, "column {} label".format(col))
            ds.set_column_unit(col, "column {} unit".format(col))

    if add_scale:
        for axis in range(len(shape) - 1):
            scale = np.random.random() * np.arange(shape[axis])
            ds.set_axis_scale(axis, scale)

    if add_attributes:
        ds.attrs["simple_int"] = 123
        ds.attrs["big_int"] = -100000000000
        ds.attrs["simple_float"] = np.pi
        ds.attrs["big_float"] = -1.0e80
        ds.attrs["simple_str"] = "hello"
        ds.attrs["tricky_str"] = "a tricky:string \\to \"encode'"

    return ds


class TestDataSet(unittest.TestCase):

    def test_01_raise_exception_on_wrong_data_type(self):
        """See that an exception is raised if data is not np.array type."""
        with self.assertRaises(TypeError):
            DataSet("empty_dataset", shape=(8, 3), data=[1.0, 2.0, 3.0])

    def test_02_raise_exception_on_wrong_shape(self):
        """See that an exception is raised if data shape is not correct."""
        with self.assertRaises(ValueError) as exc:
            DataSet("empty_dataset", shape=(8, 3), data=np.array([[1.0, 2.0, 3.0]]))

        self.assertEqual(str(exc.exception), "Data does not match specified shape")

    def test_03_raise_exception_on_wrong_data_type(self):
        """See that an exception is raised if data type is not correct."""
        with self.assertRaises(ValueError) as exc:
            DataSet("empty_dataset", shape=(1, 3), dtype=int, data=np.array([[1.0, 2.0, 3.0]]))

        self.assertEqual(str(exc.exception), "Data does not match specified data type")

    def test_04_no_shape_nor_data_specified_raises_exception(self):
        """An exception must be raised if neither shape nor data is given as input"""
        with self.assertRaises(TypeError):
            DataSet("empty_dataset", shape=None, data=None)

    def test_05_data_dimensions_not_two_or_more(self):
        """An exception must be raised if neither data dimension and shape is less than 2"""
        with self.assertRaises(ValueError) as exc:
            DataSet("empty_dataset", shape=(3,), data=np.array([1.0, 2.0, 3.0]))

        self.assertEqual(str(exc.exception), "Dataset must have at least 2 axes")

    def test_06_data_dimension_invalid(self):
        """Zero (and negative) dimension values raise an exception"""
        with self.assertRaises(ValueError) as exc:
            DataSet("empty_dataset", shape=(1, 0), data=np.array([[]]))

        self.assertEqual(str(exc.exception), "Zero-size or negative size axes are not allowed")

    def test_10_create_empty_dataset(self):
        """Create a simple, empty DataSet instance."""

        ds = DataSet("empty_dataset", shape=(8, 3))

        self.assertEqual(ds.name, "empty_dataset")
        self.assertIsInstance(ds.data, np.ndarray)
        self.assertEqual(ds.data.shape, (8, 3))
        self.assertEqual(ds.data.dtype, np.float64)
        self.assertTrue(np.all(ds.data == 0))

        self.assertEqual(ds.axis_label, [""])
        self.assertEqual(ds.axis_unit, [""])
        self.assertEqual(ds.axis_scale, [None])
        self.assertEqual(ds.column_label, ["", "", ""])
        self.assertEqual(ds.column_unit, ["", "", ""])
        self.assertEqual(ds.attrs, {})

    def test_11_create_initialized_dataset(self):
        """Create a simple dataset from existing Numpy data."""

        data = np.sqrt(np.arange(10)).reshape(5, 2)
        ds = DataSet("my_dataset", data=data)

        self.assertEqual(ds.name, "my_dataset")
        self.assertIsInstance(ds.data, np.ndarray)
        self.assertEqual(ds.data.shape, (5, 2))
        self.assertEqual(ds.data.dtype, np.float64)
        self.assertTrue(np.all(ds.data == data))

    def test_20_labels(self):
        """Setting labels."""

        ds = DataSet("my_dataset", shape=(2, 8, 3))

        ds.set_axis_label(0, "X")
        ds.set_axis_unit(0, "um")
        ds.set_axis_label(1, "Z")
        ds.set_axis_unit(1, "mm")
        ds.set_column_label(0, "power")
        ds.set_column_unit(0, "mW")
        ds.set_column_label(1, "countrate")
        ds.set_column_unit(1, "kHz")
        ds.set_column_label(2, "temperature")
        ds.set_column_unit(2, "K")

        self.assertEqual(ds.name, "my_dataset")
        self.assertEqual(ds.axis_label, ["X", "Z"])
        self.assertEqual(ds.axis_unit, ["um", "mm"])
        self.assertEqual(ds.axis_scale, [None, None])
        self.assertEqual(ds.column_label, ["power", "countrate", "temperature"])
        self.assertEqual(ds.column_unit, ["mW", "kHz", "K"])

    def test_21_axis_scale(self):
        """Setting an axis scale."""

        ds = DataSet("my_dataset", shape=(2, 8, 3))

        ds.set_axis_scale(0, [0.1, 0.2])
        ds.set_axis_scale(1, 0.25 * np.arange(8))

        self.assertEqual(ds.name, "my_dataset")
        self.assertEqual(len(ds.axis_scale), 2)
        self.assertTrue(np.all(ds.axis_scale[0] == [0.1, 0.2]))
        self.assertTrue(np.all(ds.axis_scale[1] == 0.25 * np.arange(8)))

    def test_22_invalid_label(self):
        """Setting invalid labels raises exception."""

        ds = DataSet("my_dataset", shape=(2, 8, 3))

        with self.assertRaises(TypeError):
            ds.set_axis_label("0", "X")

        with self.assertRaises(ValueError):
            ds.set_axis_label(-1, "X")

    def test_23_invalid_axis_unit(self):
        """Setting invalid axis unit raises exception."""

        ds = DataSet("my_dataset", shape=(2, 8, 3))

        with self.assertRaises(TypeError):
            ds.set_axis_unit("0", "X")

        with self.assertRaises(ValueError):
            ds.set_axis_unit(-1, "X")

    def test_24_invalid_axis_scale(self):
        """Setting invalid axis unit raises exception."""

        ds = DataSet("my_dataset", shape=(2, 8, 3))

        with self.assertRaises(TypeError):
            ds.set_axis_scale("0", [0.1, 0.2])

        with self.assertRaises(ValueError) as exc:
            ds.set_axis_scale(-1, [0.1, 0.2])

        self.assertEqual(str(exc.exception), "Invalid value for parameter 'axis'")

        with self.assertRaises(ValueError) as exc:
            ds.set_axis_scale(0, [0.1, 0.2])
            ds.set_axis_scale(1, 0.25 * np.arange(7))

        self.assertEqual(str(exc.exception), "Invalid shape for scale array")

        with self.assertRaises(ValueError) as exc:
            ds.set_axis_scale(0, [0.1, np.nan])
            ds.set_axis_scale(1, 0.25 * np.arange(8))

        self.assertEqual(str(exc.exception), "Only finite values allowed on the axis scale")

    def test_25_invalid_column_label(self):
        """Setting invalid column labels raises exception."""

        ds = DataSet("my_dataset", shape=(2, 8, 3))

        with self.assertRaises(TypeError):
            ds.set_column_label("0", "X")

        with self.assertRaises(ValueError):
            ds.set_column_label(-1, "X")

    def test_26_invalid_column_unit(self):
        """Setting invalid column unit raises exception."""

        ds = DataSet("my_dataset", shape=(2, 8, 3))

        with self.assertRaises(TypeError):
            ds.set_column_unit("0", "X")

        with self.assertRaises(ValueError):
            ds.set_column_unit(-1, "X")

    def test_30_dataset_1col(self):
        """Create a 3-dimensional, 1-column dataset and setting labels."""

        data = np.arange(10).reshape(2, 5, 1)
        ds = DataSet("my_dataset", data=data)
        ds.set_axis_label(0, "X")
        ds.set_axis_label(1, "Y")
        ds.set_column_label(0, "Z")

        self.assertEqual(ds.name, "my_dataset")
        self.assertEqual(ds.data.shape, (2, 5, 1))
        self.assertEqual(ds.axis_label, ["X", "Y"])
        self.assertEqual(ds.axis_unit, ["", ""])
        self.assertEqual(ds.axis_scale, [None, None])
        self.assertEqual(ds.column_label, ["Z"])
        self.assertEqual(ds.column_unit, [""])
        self.assertEqual(ds.attrs, {})

    def test_40_write_hdf5_simple(self):
        """Writing a simple dataset as HDF5."""

        data = 1.4142 * (np.arange(24).reshape(8, 3) - 1)
        ds = DataSet("my_dataset", data=data)
        ds.axis_label[0] = "X axis"
        ds.axis_unit[0] = "mm"
        ds.column_label = ["red", "green", "blue"]
        ds.column_unit = ["MHz", "nm", "K"]
        scale_data = 0.1 * np.arange(8)
        ds.set_axis_scale(0, scale_data)
        ds.attrs["hello"] = "world"
        ds.attrs["number"] = 2.71

        f = h5py.File("test.h5", "w", driver="core", backing_store=False)
        qmi.data.dataset.write_dataset_to_hdf5(ds, f)

        self.assertEqual(set(f.keys()), {"my_dataset", "my_dataset_axis0_scale"})
        self.assertEqual(f["my_dataset"].shape, (8, 3))
        self.assertEqual(f["my_dataset"].dtype, np.float64)
        self.assertTrue(np.all(data == f["my_dataset"]))

        self.assertEqual(f["my_dataset"].attrs["QMI_DataSet_axis0_label"], "X axis")
        self.assertEqual(f["my_dataset"].attrs["QMI_DataSet_axis0_unit"], "mm")
        self.assertEqual(f["my_dataset"].attrs["QMI_DataSet_column1_label"], "green")
        self.assertEqual(f["my_dataset"].attrs["QMI_DataSet_column1_unit"], "nm")

        self.assertEqual(f["my_dataset"].dims[0].label, "X axis")
        self.assertTrue(np.all(scale_data == f["my_dataset"].dims[0][0]))

        self.assertEqual(f["my_dataset"].attrs["hello"], "world")
        self.assertEqual(f["my_dataset"].attrs["number"], 2.71)

        f.close()

    def test_41_write_read_hdf5(self):
        """Writing and reading various datasets as HDF5."""

        f = h5py.File("test.h5", "w", driver="core", backing_store=False)
        datasets = []

        # Create datasets and write to file.
        for (name, shape, dtype) in [
                ("t1", (4, 4), np.float64),
                ("t2", (4, 4), np.float64),
                ("t3", (3, 3, 3, 3), np.float64),
                ("t4", (4, 4, 1), np.int32)]:
            ds = _internal_create_dataset(name,
                                          shape,
                                          dtype,
                                          add_labels=True,
                                          add_scale=True,
                                          add_attributes=True)
            datasets.append(ds)
            qmi.data.dataset.write_dataset_to_hdf5(ds, f)

        # Read back from file and verify.
        for ds in datasets:
            ds2 = qmi.data.dataset.read_dataset_from_hdf5(f[ds.name])
            self.assertEqual(ds2.name, ds.name)
            self.assertIsInstance(ds2.data, np.ndarray)
            self.assertEqual(ds2.data.shape, ds.data.shape)
            self.assertEqual(ds2.data.dtype, ds.data.dtype)
            self.assertTrue(np.all(ds2.data == ds.data))
            self.assertEqual(ds2.axis_label, ds.axis_label)
            self.assertEqual(ds2.axis_unit, ds.axis_unit)
            self.assertEqual(len(ds2.axis_scale), len(ds.axis_scale))
            for axis in range(len(ds.axis_scale)):
                if ds.axis_scale[axis] is None:
                    self.assertIsNone(ds2.axis_scale[axis])
                else:
                    self.assertTrue(np.all(ds2.axis_scale[axis] == ds.axis_scale[axis]))
            self.assertEqual(ds2.column_label, ds.column_label)
            self.assertEqual(ds2.column_unit, ds.column_unit)
            self.assertEqual(ds2.attrs, ds.attrs)

        f.close()

    def test_42_write_hdf5_raises_exception_on_invalid_attrs_name(self):
        """If attrs name is equal or starting with 'QMI_DataSet' or 'DIMENSION_', and exception is raised"""
        data = 1.4142 * (np.arange(24).reshape(8, 3) - 1)
        ds = DataSet("my_dataset", data=data)
        ds.axis_label[0] = "X axis"
        ds.axis_unit[0] = "mm"
        ds.column_label = ["red", "green", "blue"]
        ds.column_unit = ["MHz", "nm", "K"]
        scale_data = 0.1 * np.arange(8)
        ds.set_axis_scale(0, scale_data)
        ds.attrs["QMI_DataSet"] = "Hello"  # <-- Error
        ds.attrs["number"] = 2.71

        f = h5py.File("test.h5", "w", driver="core", backing_store=False)
        with self.assertRaises(ValueError):
            qmi.data.dataset.write_dataset_to_hdf5(ds, f)

        ds2 = DataSet("my_dataset", data=data)
        ds2.axis_label[0] = "X axis"
        ds2.axis_unit[0] = "mm"
        ds2.column_label = ["red", "green", "blue"]
        ds2.column_unit = ["MHz", "nm", "K"]
        scale_data = 0.1 * np.arange(8)
        ds2.set_axis_scale(0, scale_data)
        ds2.attrs["DIMENSION_"] = "Hello"  # <-- Error
        ds2.attrs["number"] = 2.71

        with self.assertRaises(ValueError):
            qmi.data.dataset.write_dataset_to_hdf5(ds2, f)

    def test_50_write_text_simple(self):
        """Writing a simple dataset as text."""

        data = 1.4142 * (np.arange(24).reshape(8, 3) - 1)
        ds = DataSet("my_dataset", data=data)
        ds.axis_label[0] = "X axis"
        ds.axis_unit[0] = "mm"
        ds.column_label = ["red", "green", "blue"]
        ds.column_unit = ["MHz", "nm", "K"]
        scale_data = 0.1 * np.arange(8)
        ds.set_axis_scale(0, scale_data)
        ds.attrs["hello"] = "world"
        ds.attrs["number"] = 2.71

        with io.StringIO() as f:
            qmi.data.dataset.write_dataset_to_text(ds, f)
            text = f.getvalue()

        lines = text.splitlines()
        p = 0
        self.assertEqual(lines[p], "# QMI_DataSet")
        p += 1
        self.assertEqual(lines[p], "#")
        p += 1

        attrs = {}
        while lines[p] != "#":
            self.assertTrue(lines[p].startswith("# "))
            q = lines[p].find(":")
            self.assertGreater(q, 2)
            self.assertEqual(lines[p][q:q + 2], ": ")
            attrs[lines[p][2:q]] = lines[p][q + 2:]
            p += 1
        p += 1

        rawdata = []
        while p < len(lines):
            words = lines[p].split()
            self.assertEqual(len(words), 4)
            rawdata.append([float(w) for w in words])
            p += 1
        self.assertEqual(len(data), 8)
        rawdata = np.array(rawdata)

        self.assertEqual(rawdata.shape, (8, 4))
        self.assertTrue(np.all(rawdata == np.column_stack([scale_data, data])))

        self.assertEqual(attrs["QMI_DataSet_name"], repr("my_dataset"))
        self.assertEqual(attrs["QMI_DataSet_axis0_label"], repr("X axis"))
        self.assertEqual(attrs["QMI_DataSet_axis0_unit"], repr("mm"))
        self.assertEqual(attrs["QMI_DataSet_column1_label"], repr("red"))
        self.assertEqual(attrs["QMI_DataSet_column2_label"], repr("green"))
        self.assertEqual(attrs["QMI_DataSet_column2_unit"], repr("nm"))
        self.assertEqual(attrs["hello"], repr("world"))
        self.assertEqual(attrs["number"], repr(2.71))

    def test_51_write_read_text(self):
        """Writing and reading various datasets as text."""

        # Create datasets and write to file.
        for (name, shape, dtype) in [
                ("t1", (4, 4), np.float64),
                ("t2", (4, 4), np.float64),
                ("t3", (3, 3, 3, 3), np.float64),
                ("t4", (4, 4, 1), np.int32)]:
            ds = _internal_create_dataset(name,
                                          shape,
                                          dtype,
                                          add_labels=True,
                                          add_scale=True,
                                          add_attributes=True)
            with io.StringIO() as f:
                qmi.data.dataset.write_dataset_to_text(ds, f)
                f.seek(0)
                ds2 = qmi.data.dataset.read_dataset_from_text(f)

            self.assertEqual(ds2.name, ds.name)
            self.assertIsInstance(ds2.data, np.ndarray)
            self.assertEqual(ds2.data.shape, ds.data.shape)
            self.assertTrue(np.all(ds2.data == ds.data))
            self.assertEqual(ds2.axis_label, ds.axis_label)
            self.assertEqual(ds2.axis_unit, ds.axis_unit)
            self.assertEqual(len(ds2.axis_scale), len(ds.axis_scale))
            for axis in range(len(ds.axis_scale)):
                if ds.axis_scale[axis] is None:
                    self.assertIsNone(ds2.axis_scale[axis])
                else:
                    self.assertTrue(np.all(ds2.axis_scale[axis] == ds.axis_scale[axis]))
            self.assertEqual(ds2.column_label, ds.column_label)
            self.assertEqual(ds2.column_unit, ds.column_unit)
            self.assertEqual(ds2.attrs, ds.attrs)

    def test_52_write_text_raises_exception_on_invalid_attrs_name(self):
        """If attrs name is equal or starting with 'QMI_DataSet' or 'DIMENSION_', and exception is raised"""
        data = 1.4142 * (np.arange(24).reshape(8, 3) - 1)
        ds = DataSet("my_dataset", data=data)
        ds.axis_label[0] = "X axis"
        ds.axis_unit[0] = "mm"
        ds.column_label = ["red", "green", "blue"]
        ds.column_unit = ["MHz", "nm", "K"]
        scale_data = 0.1 * np.arange(8)
        ds.set_axis_scale(0, scale_data)
        ds.attrs["QMI_DataSet"] = "Hello"  # <-- Error
        ds.attrs["number"] = 2.71

        with self.assertRaises(ValueError), io.StringIO() as f:
            qmi.data.dataset.write_dataset_to_text(ds, f)

        ds2 = DataSet("my_dataset", data=data)
        ds2.axis_label[0] = "X axis"
        ds2.axis_unit[0] = "mm"
        ds2.column_label = ["red", "green", "blue"]
        ds2.column_unit = ["MHz", "nm", "K"]
        scale_data = 0.1 * np.arange(8)
        ds2.set_axis_scale(0, scale_data)
        ds2.attrs["DIMENSION_"] = "Hello"  # <-- Error
        ds2.attrs["number"] = 2.71

        with self.assertRaises(ValueError), io.StringIO() as f:
            qmi.data.dataset.write_dataset_to_text(ds, f)

        ds2 = DataSet("my_dataset", data=data)
        ds2.axis_label[0] = "X axis"
        ds2.axis_unit[0] = "mm"
        ds2.column_label = ["red", "green", "blue"]
        ds2.column_unit = ["MHz", "nm", "K"]
        scale_data = 0.1 * np.arange(8)
        ds2.set_axis_scale(0, scale_data)
        ds2.attrs[""] = "Hello"  # <-- Error
        ds2.attrs["number"] = 2.71

        with self.assertRaises(ValueError), io.StringIO() as f:
            qmi.data.dataset.write_dataset_to_text(ds, f)

        ds2 = DataSet("my_dataset", data=data)
        ds2.axis_label[0] = "X axis"
        ds2.axis_unit[0] = "mm"
        ds2.column_label = ["red", "green", "blue"]
        ds2.column_unit = ["MHz", "nm", "K"]
        scale_data = 0.1 * np.arange(8)
        ds2.set_axis_scale(0, scale_data)
        ds2.attrs["hello:"] = "Hello"  # <-- Error
        ds2.attrs["number"] = 2.71

        with self.assertRaises(ValueError), io.StringIO() as f:
            qmi.data.dataset.write_dataset_to_text(ds, f)

    def test_60_parse_repr(self):
        """Test an internal parse function for repr-formatted values."""

        test_values = [
            # integer
            0,
            1,
            -123456789,
            (1 << 60),
            # float
            0.0,
            np.pi,
            1.0e-100,
            -1.0e100,
            -0.0,
            np.inf,
            np.nan,
            # string
            "",
            "hello",
            "multi\nline"
            "symbols: % # \\$",
            "quotes ' \" ''' \"\"\" ",
            "\x00 \x01 \r \t \b",
            "Falsches Üben von Xylophonmusik quält jeden größeren Zwerg",
            "いろはにほへとちりぬるを",
            "\u2002"
        ]

        for v in test_values:
            s = repr(v)
            w = qmi.data.dataset._parse_attribute_value(s)
            self.assertIs(type(w), type(v))
            if isinstance(v, float) and np.isnan(v):
                self.assertTrue(np.isnan(w))
            else:
                self.assertEqual(w, v)


if __name__ == "__main__":
    unittest.main()



"""Generic patcher implementations."""
from unittest.mock import patch, Mock
import tempfile

import qmi
from qmi.core.context import QMI_Context, QMI_RpcProxy


class PatcherQmiRpcProxy:

    rpc_nonblocking = Mock()  # patch("qmi.core.rpc.QMI_RpcNonBlockingProxy")

    def __init__(self, context, descriptor):
        self._context = context
        self._descriptor = descriptor

    def get_pid(self):
        return 123

    def get_version(self):
        return "SomeVersion"


class PatcherQmiContext:
    """Patches the qmi_context global variable which results in an instanced qmi_context.
    Note that only within the context of this patcher the qmi_context is a singleton.
    """

    def __init__(self, name=None):
        self.name = name
        self.qmi_patcher = None

    def start(self, context_name=None, config=None):
        """Start the qmi_context and patcher."""
        context_name = context_name or self.name
        self.qmi_patcher = patch("qmi.core.context_singleton._qmi_context", QMI_Context(context_name, config))
        self.qmi_patcher.start()
        qmi.core.context_singleton._qmi_context.start()
        qmi.core.context_singleton._connect_to_peers()

    def stop(self):
        """Stop the qmi_context and patcher."""
        qmi.stop()
        self.qmi_patcher.stop()

    def make_instrument(self, instrument_name, instrument_class, *args, **kwargs):
        return qmi.core.context_singleton._qmi_context.make_instrument(
            instrument_name, instrument_class, *args, **kwargs
        )

    def connect_to_peer(self, peer_context_name, peer_address):
        return None

    def make_peer_context_proxy(self, context_name):
        return PatcherQmiRpcProxy(self, None)

    def disconnect_from_peer(self, peer_context_name):
        return None

    def resolve_file_name(self, output_dir):
        temp_dir = tempfile.mkdtemp()
        return temp_dir

    def get_qmi_home_dir(self):
        temp_dir = tempfile.mkdtemp()
        return temp_dir

    def get_config(self):
        return Mock()



"""
Controller task for NSG.
"""

import time
from typing import NamedTuple

from qmi.core.exceptions import QMI_Exception, QMI_TaskStopException
from qmi.core.task import QMI_Task, QMI_TaskRunner
from qmi.core.pubsub import QMI_Signal
from qmi.instruments.dummy.noisy_sine_generator import NoisySineGenerator


class NoisySineGeneratorSettings(NamedTuple):
    """Noisy sine generator settings.

    Attributes:
        frequency: frequency of the sine wave.
        amplitude: amplitude of the sine wave.
        noise:     noise standard deviation.
    """
    frequency: float = 2.0
    amplitude: float = 100.0
    noise: float = 1.0


class NoisySineGeneratorStatus(NamedTuple):
    """Noisy sine generator status.

    Attributes:
        runtime: time running.
    """
    runtime: float


class NoisySineGeneratorController(QMI_Task):
    """Controller for the noisy sine generator.

    This is a simple task for controlling a noisy sine generator.
    """

    sig_sample = QMI_Signal([float, float])
    sig_status = QMI_Signal([NoisySineGeneratorStatus])

    def __init__(self,
                 task_runner: QMI_TaskRunner,
                 name: str,
                 generator: NoisySineGenerator,
                 settings: NoisySineGeneratorSettings,
                 sample_time: float = 1.0
                ) -> None:
        """Initialize the controller task.

        Parameters:
            task_runner:    Task runner instance (will be provided by QMI via `make_task()`).
            name:           Name of the task.
            generator:      Instance of NoisySineGenerator to be controlled.
            settings:       Initial settings for the NoisySineGenerator.
            sample_time:    Sample time in seconds (default: 1.0).
        """
        super().__init__(task_runner, name)

        self._nsg = generator
        self.settings: NoisySineGeneratorSettings = settings
        self._sample_time = sample_time

    def _config_nsg(self) -> None:
        self._nsg.set_frequency(self.settings.frequency)
        self._nsg.set_amplitude(self.settings.amplitude)
        self._nsg.set_noise(self.settings.noise)

    def run(self) -> None:
        """Main loop."""
        # Configure the device.
        self._config_nsg()

        # Start the sampling loop.
        t_start = time.monotonic()
        try:
            # `stop_requested()` will return True if the task should stop (e.g. when context is torn down).
            while not self.stop_requested():

                # If new settings have been pushed to the task, they are put in a FIFO. Calling `update_settings()`
                # returns True if the FIFO is non-empty. One settings object is popped from the FIFO and stored in
                # `self.settings`. The routine returns False if the FIFO is empty.
                if self.update_settings():
                    self._config_nsg()

                # Sample the NSG.
                t_current = time.monotonic() - t_start
                sample = self._nsg.get_sample()
                status = NoisySineGeneratorStatus(runtime=t_current)

                # Broadcast the sample and the status of the NSG.
                self.sig_sample.publish(t_current, sample)
                self.sig_status.publish(status)

                # Sleep until the next sample time.
                t_next = t_current + self._sample_time
                delta = t_next - time.monotonic()
                if delta > 0:
                    self.sleep(duration=delta)

        except QMI_TaskStopException:
            # Do nothing, this is expected when the task is stopped while it is sleeping.
            pass

        except QMI_Exception:
            # Unexpected exceptions are re-raised.
            raise

        finally:
            # Clean-up code goes here.
            pass



"""
Manage access to programs and data on the ADwin.
"""

import json
import logging
import os
import os.path
from dataclasses import field
from typing import Any, Dict, Iterable, List, NamedTuple, Set, Tuple, Union

import numpy as np

from qmi.core.config import load_config_file
from qmi.core.config_struct import config_struct_from_dict, configstruct
from qmi.core.exceptions import (QMI_ConfigurationException,
                                 QMI_InstrumentException)
from qmi.instruments.adwin.adwin import Adwin_Base
from qmi.utils.adbasic_compiler import check_compiler_result, compile_program
from qmi.utils.adbasic_parser import (ArrayElemDesc, FParDesc, ParameterDesc,
                                      ParameterInfo, ParDesc,
                                      analyze_parameter_info,
                                      parse_adbasic_program)

# Global variable holding the logger for this module.
_logger = logging.getLogger(__name__)


@configstruct
class CfgAdwinProgram:
    """Configuration for a specific Adwin program.

    Parameters:
        file:       File name of the top-level ".bas" file without the ".bas" extension.
                    This file name is interpreted relative to the Adwin program directory.
        slot:       Process ("slot") number in the ADwin (range 1 to 10). Not the same as physical encasing slots.
        trigger:    Event trigger source, either "timer" or "external".
        priority:   Process priority, either an integer between -10 and +10 for a low priority process,
                    or 1000 for a high priority process.
        parse_parameters: True to discover parameters by parsing the ADbasic source code.
        par:        Dictionary mapping parameter name to global Par_nn index.
        fpar:       Dictionary mapping parameter name to global FPar_nn index.
        data:       Dictionary mapping array name to global Data_nnn index.
        par_array:  Dictionary mapping parameter name to list [nnn, k] referring to array element Data_nnn[k].
    """
    file:       str
    slot:       int
    trigger:    str
    priority:   int
    parse_parameters: bool = False
    par:        Dict[str, int] = field(default_factory=dict)
    fpar:       Dict[str, int] = field(default_factory=dict)
    data:       Dict[str, int] = field(default_factory=dict)
    par_array:  Dict[str, Tuple[int, int]] = field(default_factory=dict)  # Tuple does not get parsed correctly in JSON!


class ProgramInfo(NamedTuple):
    """Information about an ADwin program.

    Attributes:
        file:       Path name to the top-level ".bas" file without the ".bas" extension.
        slot:       Process slot number in the ADwin (range 1 to 10). Not the same as physical encasing slots.
        trigger:    Event trigger source, either "timer" or "external".
        priority:   Process priority, either an integer between -10 and +10 for a low priority process,
                    or 1000 for a high priority process.
        parameters: Information about parameters and data arrays in the ADbasic program.
    """
    file: str
    slot: int
    trigger: str
    priority: int
    param_info: ParameterInfo

    @classmethod
    def from_config(cls, config: CfgAdwinProgram, program_dir: str) -> 'ProgramInfo':
        """Create a ProgramInfo from a CfgAdwinProgram instance.

        Parameters:
            config: Adwin program configuration.
            program_dir: directory where adwin program files are located.

        Returns:
            Instance of ProgramInfo.

        Raises:
            QMI_ConfigurationException: If an error occurs while reading parameter configuration data.
            ParseException: If an error occurs while parsing the ADbasic source code.
        """
        filename = os.path.join(program_dir, config.file)

        if config.parse_parameters:
            # Get parameter info by parsing the ADbasic source code.
            symbols = parse_adbasic_program(f"{filename}.bas", program_dir)
            param_info = analyze_parameter_info(symbols)

        else:
            # Use explicitly specified parameters.
            param_items: List[Tuple[str, ParameterDesc]] = []
            for (name, par_index) in config.par.items():
                param_items.append((name, ParDesc(par_index)))
            for (name, fpar_index) in config.fpar.items():
                param_items.append((name, FParDesc(fpar_index)))
            for (name, ref) in config.par_array.items():
                (data_index, elem_index) = ref
                param_items.append((name, ArrayElemDesc(data_index, elem_index)))

            param: Dict[str, ParameterDesc] = {}
            for (name, desc) in param_items:
                if name in param:
                    raise QMI_ConfigurationException("Duplicate use of parameter name {!r} in ADwin program \"{!r}\"!"
                                                     .format(name, config.file))
                param[name] = desc

            param_info = ParameterInfo(param=param, data=config.data)

        # Create program information structure.
        return cls(file=filename,
                   slot=config.slot,
                   trigger=config.trigger.lower(),
                   priority=config.priority,
                   param_info=param_info)

    @classmethod
    def from_config_file(cls, config_file: str, program_dir: str) -> 'ProgramInfo':
        """Create a ProgramInfo from a program configurations file.

        Parameters:
            config_file: A configuration file with Adwin program configuration(s).
            program_dir: directory where adwin program files are located.

        Returns:
            Instance of ProgramInfo.
        """
        config_dict = load_config_file(config_file)
        config = config_struct_from_dict(config_dict, CfgAdwinProgram)
        return cls.from_config(config, program_dir)


class AdwinProgramLibrary:
    """Manage a collection of ADwin programs."""

    def __init__(
            self,
            program_dir: str,
            config_dir: str,
            processor_type: str = "T12",
            hardware_type: str = "PII"
    ) -> None:
        """Initialize a collection of ADwin programs.

        Parameters:
            program_dir:     Directory where ADbasic source code and compiled programs are located.
                             File names of programs are interpreted relative to this directory.
            config_dir:      Directory where program-specific configuration files are located.
            processor_type:  Parameter to select a specific processor type "T<x>", where <x> is 11, 12 or 12.1.
                             Default is 'T12'
            hardware_type:   Parameter to select a specific processor type "<x>", where <x> is P, PII, G, GII
                             (P=pro, G=Gold). Default is 'PII'.
        """
        self._program_dir = program_dir
        self._config_dir = config_dir
        self._processor_type = processor_type
        self._hardware_type = hardware_type
        self._program_config: Dict[str, CfgAdwinProgram] = {}
        self._program_info_cache: Dict[str, ProgramInfo] = {}
        self._read_program_config_files()

    def _read_program_config_files(self) -> None:
        """Scan the configuration directory and read program-specific configuration files."""

        file_names = os.listdir(self._config_dir)

        for file_name in file_names:
            if not file_name.startswith("."):
                (prog_name, ext) = os.path.splitext(file_name)
                if (ext == ".conf") and (not prog_name.startswith("_")):
                    _logger.debug("Reading Adwin program config %s", file_name)
                    full_name = os.path.join(self._config_dir, file_name)
                    try:
                        cfg_dict = load_config_file(full_name)
                        prog_cfg = config_struct_from_dict(cfg_dict, CfgAdwinProgram)
                    except (json.JSONDecodeError, QMI_ConfigurationException) as exc:
                        raise QMI_ConfigurationException(
                            "Error reading Adwin program config {}".format(file_name)) from exc
                    self._check_program_config(prog_name, prog_cfg)
                    self._program_config[prog_name] = prog_cfg

    @staticmethod
    def _check_program_config(prog_name: str, prog_cfg: CfgAdwinProgram) -> None:

        if prog_cfg.slot < 1 or prog_cfg.slot > 10:
            raise QMI_ConfigurationException("Invalid ADwin process slot number {} for program {!r}"
                                             .format(prog_cfg.slot, prog_name))

        if prog_cfg.trigger.lower() not in ("timer", "external"):
            raise QMI_ConfigurationException("Invalid ADwin process trigger {!r} for program {!r}"
                                             .format(prog_cfg.trigger, prog_name))

        if ((prog_cfg.priority < -10) or (prog_cfg.priority > 10)) and (prog_cfg.priority != 1000):
            raise QMI_ConfigurationException("Invalid ADwin process priority {} for program {!r}".
                                             format(prog_cfg.priority, prog_name))

        if prog_cfg.parse_parameters and (prog_cfg.par or prog_cfg.fpar or prog_cfg.data or prog_cfg.par_array):
            raise QMI_ConfigurationException("Invalid configuration for ADwin program {!r}".format(prog_name)
                                             + ": Can not parse source code and specify explicit parameters")

    def list_programs(self) -> List[str]:
        """Return a list of available programs."""
        return list(self._program_config.keys())

    def get_program_info(self, program_name: str) -> ProgramInfo:
        """Return information about the specified program.

        This function may read a program-specific configuration file or
        parse the ADbasic source code to obtain information about the
        parameters and data arrays used by this program.

        Parameters:
            program_name: Unique name of the program, used to identify
                          this program within the ADwin configuration data.

        Returns:
            Information about the specified program.

        Raises:
            KeyError: If the specified program name is unknown.
            QMI_ConfigurationException: If an error occurs while reading parameter configuration data.
            ParseException: If an error occurs while parsing the ADbasic source code.
        """

        # Return cached information if available.
        info = self._program_info_cache.get(program_name)
        if info is not None:
            return info

        # Create program info from config
        info = ProgramInfo.from_config(self._program_config[program_name], self._program_dir)

        # Add to cached information
        self._program_info_cache[program_name] = info

        return info

    def compile(self, program_name: str) -> None:
        """Compile the specified program with the ADbasic compiler.

        Parameters:
            program_name: Unique name of the program.

        Raises:
            KeyError: If the specified program name is unknown.
            AdbasicCompilerException: If an error occurs while compiling the program.
        """

        # Look up in program table.
        prog_cfg = self._program_config[program_name]

        _logger.info("Compiling ADbasic program %s", program_name)

        # Run the compiler in the directory that contains the ".bas" file.
        # This implies that include files are resolved relative to the location of the .bas file.
        working_dir = os.path.join(self._program_dir, os.path.dirname(prog_cfg.file))
        basic_filename = os.path.basename(prog_cfg.file) + ".bas"

        # Run compiler.
        result = compile_program(basic_filename,
                                 processor_type=self._processor_type,
                                 hardware_type=self._hardware_type,
                                 trigger=prog_cfg.trigger,
                                 process_number=prog_cfg.slot,
                                 priority=prog_cfg.priority,
                                 working_dir=working_dir,
                                 pretty_print=False)

        # Raise exception if compiler failed.
        check_compiler_result(basic_filename, result)


class AdwinProcess:
    """Access a specific process running on a specific ADwin.

    An instance of this class represents a specific process on a specific
    ADwin instrument.  The process may be in one of several states:
    running, loaded but not yet running, or not yet loaded.

    If the process is loaded (and possibly running), data may be transferred
    from Python to the ADwin process and vice versa.

    Global parameters and data arrays are accessed by name. These names are
    mapped to `Par_nn`, `FPar_nn` and `Data_nn` indices by lookup in a
    name table. This table is either specified explicitly as part of
    the ADwin configuration, or derived from parsing the ADbasic code.
    """

    def __init__(self, adwin: Adwin_Base, name: str, program_info: ProgramInfo) -> None:
        """Initialize this object instance.

        This does not yet access the ADwin and/or start the process.

        Parameters:
            adwin:      Adwin instrument driver.
            name:       Program name.
            program_info: Information about the ADwin program.
        """
        self._name = name
        self._adwin = adwin
        self._file_name = program_info.file
        self._process_number = program_info.slot
        self._param_info = program_info.param_info
        self._processor_type = adwin.get_processor_type()

    @staticmethod
    def _get_dict_item_case_insensitive(key: str, dict_items: Dict[str, Any]):
        for dict_item_key in dict_items.keys():
            if key.lower() == dict_item_key.lower():
                return dict_items[dict_item_key]
        raise KeyError("Key not found!")

    def _get_par_desc(self, par_name):
        try:
            return self._get_dict_item_case_insensitive(par_name, self._param_info.param)
        except KeyError:
            raise ValueError("Unknown parameter {} for process {}".format(par_name, self._name))

    def _get_data_idx(self, data_name):
        try:
            return self._get_dict_item_case_insensitive(data_name, self._param_info.data)
        except KeyError:
            raise ValueError("Unknown data array {} for process {}".format(data_name, self._name))

    def load(self) -> None:
        """Load the Adwin process but do not start it yet.

        If a process is already running in the same process slot number, that process will be stopped.
        """
        if "T12" in self._processor_type.upper():
            bin_file = self._file_name + ".TC{}".format(self._process_number % 10)  # T12 and T12.1

        else:
            bin_file = self._file_name + ".TB{}".format(self._process_number % 10)  # T11 gets extension "TB{}".

        self._adwin.stop_process(self._process_number)
        self._adwin.wait_for_process(self._process_number, Adwin_Base.PROCESS_STOP_TIMEOUT)
        self._adwin.load_process(bin_file)

    def start(self) -> None:
        """Start running the Adwin process.

        The process must already have been loaded by calling the `load()` method.
        The process must not already be running.

        This function will not initialize or modify any parameters before starting the process.

        Raise:
            QMI_InstrumentException: If the process is already running.
        """
        if self._adwin.is_process_running(self._process_number):
            raise QMI_InstrumentException("Can not start Adwin process {}, process {} already running"
                                          .format(self._name, self._process_number))
        self._adwin.start_process(self._process_number)

    def start_with_params(self, **kwargs: Any) -> None:
        """Send parameter values to the Adwin, then start running the Adwin process.

        The process must already have been loaded by calling the `load()` method.
        The process must not already be running.

        Parameters for the Adwin process may be specified as keyword parameters to this function.
        These parameters will be sent to the Adwin prior to starting the process.

        Notes:
            Parameter name lookup is performed case-insensitive.

        Raise:
            QMI_InstrumentException: If the process is already running.
        """

        # Check that process is not running.
        if self._adwin.is_process_running(self._process_number):
            raise QMI_InstrumentException("Can not start Adwin process {}, process {} already running"
                                          .format(self._name, self._process_number))

        # Set parameters. Fill-in zero for unspecified parameters.
        params = {}
        for par_name in self._param_info.param.keys():
            try:
                params[par_name] = self._get_dict_item_case_insensitive(par_name, kwargs)
            except KeyError:
                params[par_name] = 0
        self.set_par_multiple(params)

        # Start process.
        self._adwin.start_process(self._process_number)

    def stop(self) -> None:
        """Stop running the Adwin process."""
        self._adwin.stop_process(self._process_number)
        self._adwin.wait_for_process(self._process_number, Adwin_Base.PROCESS_STOP_TIMEOUT)

    def is_running(self) -> bool:
        """Return True if the Adwin process is currently running.

        It is assumed that the process is already loaded. This function only
        checks whether any process is running in the specified slot number.
        It can not guarantee that the running process matches the expected program.
        """
        return self._adwin.is_process_running(self._process_number)

    def wait_for_process(self, timeout: float) -> None:
        """Wait until the Adwin process has stopped.

        Parameters:
            timeout: Maximum time to wait (seconds).

        Raises:
            ~qmi.core.exceptions.QMI_TimeoutException: If the process does not stop before timeout occurs.
        """
        self._adwin.wait_for_process(self._process_number, timeout)

    def get_par(self, par_name: str) -> Union[int, float]:
        """Return the current value of the specified parameter.

        Notes:
            Parameter name lookup is performed case-insensitive.

        Parameters:
            par_name: Parameter name.

        Returns:
            Value of the parameter.
        """
        param_desc = self._get_par_desc(par_name)
        if isinstance(param_desc, ParDesc):
            return self._adwin.get_par(param_desc.par_index)
        elif isinstance(param_desc, FParDesc):
            return self._adwin.get_fpar(param_desc.fpar_index)
        elif isinstance(param_desc, ArrayElemDesc):
            elems = self._adwin.get_data(param_desc.data_index, param_desc.elem_index, 1)
            return elems[0]
        else:
            raise ValueError("Unknown parameter {} for process {}".format(par_name, self._name))

    def set_par(self, par_name: str, value: Union[int, float]) -> None:
        """Change the value of the specified parameter.

        Notes:
            Parameter name lookup is performed case-insensitive.

        Parameters:
            par_name: Parameter name.
            value: New value to write to the parameter.
        """
        param_desc = self._get_par_desc(par_name)
        if isinstance(param_desc, ParDesc):
            if not isinstance(value, int):
                raise TypeError("Expecting integer value for parameter {} but got {}".format(par_name, value))
            self._adwin.set_par(param_desc.par_index, value)
        elif isinstance(param_desc, FParDesc):
            self._adwin.set_fpar(param_desc.fpar_index, value)
        elif isinstance(param_desc, ArrayElemDesc):
            elems = np.array([value])
            self._adwin.set_data(param_desc.data_index, param_desc.elem_index, elems)
        else:
            raise ValueError("Unknown parameter {} for process {}".format(par_name, self._name))

    @staticmethod
    def _find_sequential_ranges(seq: Iterable[int]) -> List[Tuple[int, int]]:
        """Find sequential ranges of in a sequence of integers."""
        sorted_seq = sorted(seq)
        ret = []
        if sorted_seq:
            start_range = sorted_seq[0]
            end_range = sorted_seq[0]
            for value in sorted_seq[1:]:
                if value == end_range + 1:
                    end_range = value
                elif value > end_range:
                    ret.append((start_range, end_range))
                    start_range = value
                    end_range = value
            ret.append((start_range, end_range))
        return ret

    def get_par_multiple(self, params: List[str]) -> Dict[str, Union[int, float]]:
        """Fetch multiple parameter values from the Adwin.

        Notes:
            Parameter name lookup is performed case-insensitive.
        """

        # Initialize result.
        result: Dict[str, Union[int, float]] = {}

        # Initialize an index of array elements to be fetched.
        params_data: Dict[int, Dict[int, str]] = {}

        # Fetch Par_nn and FPar_nn parameters.
        # Also build the index of array elements to be fetched.
        for par_name in params:
            param_desc = self._get_par_desc(par_name)
            if isinstance(param_desc, ParDesc):
                result[par_name] = self._adwin.get_par(param_desc.par_index)
            elif isinstance(param_desc, FParDesc):
                result[par_name] = self._adwin.get_fpar(param_desc.fpar_index)
            elif isinstance(param_desc, ArrayElemDesc):
                if param_desc.data_index not in params_data:
                    params_data[param_desc.data_index] = {}
                params_data[param_desc.data_index][param_desc.elem_index] = par_name
            else:
                raise ValueError("Unknown parameter {} for process {}".format(par_name, self._name))

        # Fetch array parameters.
        for data_index in sorted(params_data.keys()):
            data_elems = params_data[data_index]
            # Merge sequential index ranges within array.
            for (elem_index_start, elem_index_end) in self._find_sequential_ranges(data_elems.keys()):
                num_elems = elem_index_end + 1 - elem_index_start
                values = self._adwin.get_data(data_index, elem_index_start, num_elems)
                for k in range(num_elems):
                    par_name = data_elems[elem_index_start + k]
                    result[par_name] = values[k]

        return result

    def set_par_multiple(self, params: Dict[str, Union[int, float]]) -> None:
        """Send multiple parameter values to the Adwin.

        Notes:
            Parameter name lookup is performed case-insensitive.
        """

        # Initialize an index of array elements to be uploaded.
        params_data: Dict[int, Dict[int, Union[int, float]]] = {}

        # Upload Par_nn and FPar_nn parameters.
        # Also fill the index of array elements to be uploaded.
        for (par_name, par_value) in params.items():
            param_desc = self._get_par_desc(par_name)
            if isinstance(param_desc, ParDesc):
                if not isinstance(par_value, int):
                    raise TypeError("Expecting integer value for parameter {} but got {}".format(par_name, par_value))
                self._adwin.set_par(param_desc.par_index, par_value)
            elif isinstance(param_desc, FParDesc):
                self._adwin.set_fpar(param_desc.fpar_index, par_value)
            elif isinstance(param_desc, ArrayElemDesc):
                if param_desc.data_index not in params_data:
                    params_data[param_desc.data_index] = {}
                params_data[param_desc.data_index][param_desc.elem_index] = par_value
            else:
                raise ValueError("Unknown parameter {} for process {}".format(par_name, self._name))

        # Upload array parameters.
        for data_index in sorted(params_data.keys()):
            data_elems = params_data[data_index]
            # Merge sequential index ranges within array.
            for (elem_index_start, elem_index_end) in self._find_sequential_ranges(data_elems.keys()):
                values = [data_elems[i] for i in range(elem_index_start, elem_index_end + 1)]
                self._adwin.set_data(data_index, elem_index_start, np.array(values))

    def get_data_length(self, data_name: str) -> int:
        """Return the length of the specified global Data array.

        Notes:
            Data name lookup is performed case-insensitive.

        Parameters:
            data_name: Name of the data array.

        Returns:
            Number of elements in the array.
        """
        data_idx = self._get_data_idx(data_name)
        return self._adwin.get_data_length(data_idx)

    def get_data(self, data_name: str, first_index: int, count: int) -> np.ndarray:
        """Read values from a global DATA array.

        Notes:
            Data name lookup is performed case-insensitive.

        Parameters:
            data_name: Name of the data array.
            first_index: First index into the data array (>= 1).
            count: Number of elements to read from the data array.

        Returns:
            1D Numpy array with data values.
        """
        data_idx = self._get_data_idx(data_name)
        return self._adwin.get_data(data_idx, first_index, count)

    def get_full_data(self, data_name: str) -> np.ndarray:
        """Read all elements from a global Data array.

        Notes:
            Data name lookup is performed case-insensitive.

        Parameters:
            data_name: Name of the data array.

        Returns:
            1D Numpy array with data values.
        """
        data_idx = self._get_data_idx(data_name)
        return self._adwin.get_full_data(data_idx)

    def set_data(self, data_name: str, first_index: int, value: Union[np.ndarray, List[int], List[float]]) -> None:
        """Write values to a global Data array.

        Notes:
            Data name lookup is performed case-insensitive.

        Parameters:
            data_name: Name of the data array.
            first_index: First index into the data array (>= 1).
            value: List of values to write to the data array.
        """
        data_idx = self._get_data_idx(data_name)
        self._adwin.set_data(data_idx, first_index, value)

    def get_fifo_filled(self, data_name: str) -> int:
        """Return the number of values waiting in the FIFO.

        Notes:
            Data name lookup is performed case-insensitive.

        Parameters:
            data_name: Name of the data array (to look up in the Adwin configuration).

        Returns:
            Number of FIFO entries currently in use.
        """
        data_idx = self._get_data_idx(data_name)
        return self._adwin.get_fifo_filled(data_idx)

    def get_fifo_room(self, data_name: str) -> int:
        """Return the number of values that can be added to the FIFO without overflowing it.

        Notes:
            Data name lookup is performed case-insensitive.

        Parameters:
            data_name: Name of the FIFO data array.

        Returns:
            Number of empty FIFO entries available.
        """
        data_idx = self._get_data_idx(data_name)
        return self._adwin.get_fifo_room(data_idx)

    def read_fifo(self, data_name: str, count: int) -> np.ndarray:
        """Read data from a global FIFO variable.

        Notes:
            Data name lookup is performed case-insensitive.

        Parameters:
            data_name: Name of the FIFO data array.
            count: Number of values to read from the FIFO.

        Returns:
            1D Numpy array containing the FIFO elements.
        """
        data_idx = self._get_data_idx(data_name)
        return self._adwin.read_fifo(data_idx, count)

    def write_fifo(self, data_name: str, value: Union[np.ndarray, List[int], List[float]]) -> None:
        """Write data to a global FIFO variable.

        Notes:
            Data name lookup is performed case-insensitive.

        Parameters:
            data_name: Name of the FIFO data array.
            value: Array of values to write to the FIFO.
        """
        data_idx = self._get_data_idx(data_name)
        self._adwin.write_fifo(data_idx, value)


class AdwinManager:
    """Provide convenient access to ADwin processes.

    An instance of this class manages the processes running on a specific ADwin instrument.
    """

    def __init__(self,
                 adwin: Adwin_Base,
                 program_library: AdwinProgramLibrary,
                 auto_load_programs: List[str]
                 ) -> None:
        self._adwin = adwin
        self._program_library = program_library
        self._auto_load_programs = auto_load_programs
        self._check_auto_load_processnrs()

    def _check_auto_load_processnrs(self) -> None:
        """Sanity check for process number conflicts among automatically loaded programs."""
        numbers_used: Set[int] = set()
        program_names = set(self._program_library.list_programs())
        for program_name in self._auto_load_programs:
            if program_name not in program_names:
                raise ValueError("Unknown program {!r} in auto_load_programs".format(program_name))
            program_info = self._program_library.get_program_info(program_name)
            if program_info.slot in numbers_used:
                raise ValueError("Duplicate auto_load program in process slot number {} ({})"
                                 .format(program_info.slot, program_name))
            numbers_used.add(program_info.slot)

    def get_adwin(self) -> Adwin_Base:
        """Return the Adwin instrument driver used by this manager."""
        return self._adwin

    def get_program_library(self) -> AdwinProgramLibrary:
        """Return the Adwin program library used by this manager."""
        return self._program_library

    def reboot(self) -> None:
        """Reboot the Adwin system, clear data and restart programs.

        This function must be called after power-on, before communication
        with the Adwin is possible. This function may be called again
        to perform a full reset.

        This function boots the Adwin, removes all processes and data.
        It then loads the default processes as specified in the Adwin configuration.
        """
        self._adwin.reboot()
        self._adwin.check_ready()

        # Reload standard processes.
        for program_name in self._auto_load_programs:
            adwin_process = self.get_process(program_name)
            adwin_process.load()

    def list_programs(self) -> List[str]:
        """Return a list of available programs."""
        return self._program_library.list_programs()

    def get_process(self, program_name: str) -> AdwinProcess:
        """Get an `AdwinProcess` instance for the specified program."""
        program_info = self._program_library.get_program_info(program_name)
        return AdwinProcess(self._adwin, program_name, program_info)



"""
Dummy QMI instrument class for demoing and tutorial examples.
"""
# Alternative, QMI naming convention approved name
from qmi.instruments.dummy.noisy_sine_generator import NoisySineGenerator as Dummy_NoisySineGenerator



"""
Instrument driver for the Wieserlabs FlexDDS-NG Dual signal source.
"""

import enum
import logging
import re
from typing import NamedTuple, Optional

from qmi.core.context import QMI_Context
from qmi.core.exceptions import QMI_InstrumentException
from qmi.core.instrument import QMI_Instrument
from qmi.core.rpc import rpc_method
from qmi.core.transport import create_transport


# Global variable holding the logger for this module.
_logger = logging.getLogger(__name__)


class OutputChannel(enum.IntEnum):
    """FlexDDS commands can be specified to apply to one or both of the output channels."""
    OUT0 = 0
    OUT1 = 1
    BOTH = 2


class DdsRegister(enum.IntEnum):
    """Register names and addresses of the AD9910 DDS."""
    CFR1 = 0x00
    CFR2 = 0x01
    ADAC = 0x03
    IOUR = 0x04
    FTW  = 0x07
    POW  = 0x08
    ASF  = 0x09
    DRL  = 0x0B
    DRSS = 0x0C
    DRR  = 0x0D
    STP0 = 0x0E
    STP1 = 0x0F
    STP2 = 0x10
    STP3 = 0x11
    STP4 = 0x12
    STP5 = 0x13
    STP6 = 0x14
    STP7 = 0x15
    RAMB = 0x16
    RAM32E = 0x17
    RAM64C = 0x18
    RAM64E = 0x19


class DcpRegister(enum.IntEnum):
    """Register names and addresses of the DCP."""
    CFG_BNC_A   = 0x080
    CFG_BNC_B   = 0x081
    CFG_BNC_C   = 0x082
    CFG_UPDATE  = 0x084
    CFG_OSK     = 0x085
    CFG_DRCTL   = 0x086
    CFG_DRHOLD  = 0x087
    CFG_PROFILE = 0x088
    AM_S0       = 0x100
    AM_S1       = 0x101
    AM_O        = 0x102
    AM_O0       = 0x103
    AM_O1       = 0x104
    AM_CFG      = 0x105


class PllStatus(NamedTuple):
    """PLL status of the Wieserlabs FlexDDS."""
    pll1_lock:      bool
    pll2_lock:      bool
    holdover:       bool
    clkin0_lost:    bool
    clkin1_lost:    bool


class Wieserlabs_FlexDDS_NG_Dual(QMI_Instrument):
    """Instrument driver for the Wieserlabs FlexDDS-NG Dual signal source.

    The instrument is based on two AD9910 direct digital synthesizer (DDS)
    chips, one for each output channel. The instrument also contains one
    DDS Command Processor (DCP) for each channel. The DCP can be used to
    program the DDS and to enable analog modulation modes.

    This driver provides a simple, limited, high-level interface to the instrument
    through the methods `set_single_tone()`, `set_amplitude_modulation()` and
    `set_frequency_modulation()`. These methods support only a subset of
    the features of the instrument.

    This driver also provides a low-level interface to the instrument through
    the the `dcp_XXX()` methods. These methods support more functionality
    (however waiting, triggering and digital I/O are still not supported).
    Detailed understanding of the FlexDDS user manual and the AD9910 datasheet
    are needed to use the low-level interface.
    """

    # Response timeout for normal commands.
    COMMAND_RESPONSE_TIMEOUT = 2.0

    # Frequency resolution in Hz.
    # Only integer multiples of this frequency can be represented exactly.
    FREQUENCY_RESOLUTION = 1.0e9 / 2**32

    # Max 400 MHz.
    MAX_FREQUENCY = 400.0e6

    # Maximum analog input level in Volt.
    MAX_ANALOG_INPUT_VOLT = 0.5

    def __init__(self,
                 context: QMI_Context,
                 name: str,
                 transport: str
                 ) -> None:
        """Initialize the instrument driver.

        Arguments:
            name:       Name for this instrument instance.
            transport:  QMI transport descriptor to connect to the instrument.
                        The transport descriptor will typically specify a
                        serial port and baud rate, for example
                        "serial:/dev/ttyACM0:baudrate=115200".
        """
        super().__init__(context, name)
        self._transport = create_transport(transport, default_attributes={"baudrate": 115200})

    def _write(self, cmd: str) -> None:
        """Send a command to the instrument."""
        self._transport.write(cmd.encode("ascii") + b"\r")

    @rpc_method
    def open(self) -> None:
        _logger.info("[%s] Opening connection to instrument", self._name)
        self._transport.open()

        try:
            # Discard any pending data from the instrument.
            self._transport.discard_read()

            # By default the FlexDDS starts in "interactive" mode. This means
            # it sends interactive prompts, command echos, and log messages.
            # Switch it to non-interactive mode to get only meaningful responses.
            self._write("")  # flush potential partial command string
            self._write("interactive off")

            # Discard response up to the point where interactive mode stops.
            _resp = self._transport.read_until(b"Interactive off\r\n", timeout=self.COMMAND_RESPONSE_TIMEOUT)

        except Exception:
            self._transport.close()
            raise

        super().open()

    @rpc_method
    def close(self) -> None:
        _logger.info("[%s] Closing connection to instrument", self._name)
        super().close()
        self._transport.close()

    @rpc_method
    def get_version(self) -> str:
        """Read the instrument version string."""
        self._transport.discard_read()  # Discard potential replies from previous commands.
        self._write("version")

        # The reply format to the "version" command is not very well specified.
        # It typically starts with 'Version: "version_string" (other stuff...)'
        # and ends with 'USER: ...'.
        version_string = None
        while True:
            resp = self._transport.read_until(b"\n", self.COMMAND_RESPONSE_TIMEOUT)
            resp_str = resp.decode("latin1")
            version_match = re.match(r'^Version:\s*"([^"]+)"', resp_str)
            if version_match:
                version_string = version_match.group(1)
            if resp_str.find("USER:") >= 0:
                break  # this was the last line of the reply

        if version_string is None:
            raise QMI_InstrumentException("Unexpected reply to 'version' command")

        return version_string

    @rpc_method
    def get_pll_status(self) -> PllStatus:
        """Read the PLL status.

        This is an undocumented command.
        The meaning of the result is not fully understood, however it
        seems that the attribute `pll1_lock` shows whether the 10 MHz
        external reference is detected.
        """
        self._transport.discard_read()  # Discard potential replies from previous commands.
        self._write("lmkpll")

        # Read results.
        # The reply format is not fully specified, but generally looks something like this:
        #
        # LMK PLL status: rv=666
        #   DLD (lock detect): PLL1: *UNLOCK*   PLL2: locked      BOTH: *UNLOCK*
        #   Holdover . . . . : *ACTIVE*
        #   DAC. . . . . . . : low
        #   LOS (signal loss): CLKIN0: *LOST*       CLKIN1: not lost
        #   CLKIN selected . : CLKIN0: *SELECT*     CLKIN1: (unsel)
        #
        pll1_lock: Optional[bool] = None
        pll2_lock: Optional[bool] = None
        holdover: Optional[bool] = None
        clkin0_lost: Optional[bool] = None
        clkin1_lost: Optional[bool] = None
        while True:
            resp = self._transport.read_until(b"\n", self.COMMAND_RESPONSE_TIMEOUT)
            resp_str = resp.decode("latin1")
            lock_match = re.match(r"^\s*DLD.*PLL1:\s*(\*UNLOCK\*|locked)\s*PLL2:\s*(\*UNLOCK\*|locked)", resp_str)
            if lock_match:
                pll1_lock = (lock_match.group(1) == "locked")
                pll2_lock = (lock_match.group(2) == "locked")
            holdover_match = re.match(r"^\s*Holdover.*:\s*(\*ACTIVE\*|inact)", resp_str)
            if holdover_match:
                holdover = (holdover_match.group(1) == "*ACTIVE*")
            loss_match = re.match(r"^\s*LOS.*CLKIN0:\s*(\*LOST\*|not lost)\s*CLKIN1:\s*(\*LOST\*|not lost)", resp_str)
            if loss_match:
                clkin0_lost = (loss_match.group(1) == "*LOST*")
                clkin1_lost = (loss_match.group(2) == "*LOST*")
            if resp_str.find("CLKIN selected") >= 0:
                break  # this was the last line of the reply

        if ((pll1_lock is None) or (pll2_lock is None) or (holdover is None)
                or (clkin0_lost is None) or (clkin1_lost is None)):
            raise QMI_InstrumentException("Unexpected reply to 'lmkpll' command")

        return PllStatus(pll1_lock=pll1_lock,
                         pll2_lock=pll2_lock,
                         holdover=holdover,
                         clkin0_lost=clkin0_lost,
                         clkin1_lost=clkin1_lost)

    @staticmethod
    def _format_channel(channel: OutputChannel) -> str:
        """Return the output channel selector in the format expected by the instrument."""
        if channel == OutputChannel.OUT0:
            return "0"
        elif channel == OutputChannel.OUT1:
            return "1"
        else:
            return ""  # omit channel number to select both channels

    @staticmethod
    def _calc_reg_cfr1(invsinc_filter: bool = True,
                       sine_output: bool = True,
                       osk_enable: bool = False,
                       manual_osk: bool = False,
                       auto_osk: bool = False
                       ) -> int:
        """Calculate a value for the CFR1 register of the AD9910 DDS.

        Parameters:
            invsinc_filter:   Enable inverse Sinc filter.
            sine_output:      True to select sine output; False to select cosine output.
            osk_enable:       Enable OSK mode.
            manual_osk:       Enable manual OSK via external OSK pin (when osk_enable is also True).
            auto_osk:         Enable automatic OSK (when osk_enable is also True).
        """

        # Bits 7..0 are not writable.
        # We set these to their default values.
        value = 0x00000002

        if manual_osk:
            value |= (1 << 23)
        if invsinc_filter:
            value |= (1 << 22)
        if sine_output:
            value |= (1 << 16)
        if osk_enable:
            value |= (1 << 9)
        if auto_osk:
            value |= (1 << 8)

        return value

    @staticmethod
    def _calc_reg_cfr2(ampl_scale_en: bool = False,
                       match_latency: bool = True,
                       hold_last_value: bool = True,
                       parallel_data_en: bool = False,
                       fm_gain: int = 0
                       ) -> int:
        """Calculate a value for the CFR2 register of the AD9910 DDS.

        Parameters:
            ampl_scale_en:    Enable amplitude scaling (if False, amplitudes settings are ignored)
            match_latency:    Simultaneously apply changes to amplitude, phase and frequency.
            hold_last_value:  Hold last value of the parallel data port when the DCP does not send a new word.
            parallel_data_en: Enable the parallel data port (analog modulation modes).
            fm_gain:          Shift count for frequency modulation data.
        """

        fm_gain = int(fm_gain)
        if fm_gain < 0 or fm_gain > 15:
            raise ValueError("Invalid shift count")

        # Bits 23..22, bits 11..9 and bit 5 are not writable.
        # We set these to their default values.
        value = 0x00400800

        if ampl_scale_en:
            value |= (1 << 24)
        if match_latency:
            value |= (1 << 7)
        if hold_last_value:
            value |= (1 << 6)
        if parallel_data_en:
            value |= (1 << 4)
        value |= (fm_gain & 15)

        return value

    @staticmethod
    def _calc_reg_stp(ampl_scale: int, phase_offset: int, ftw: int) -> int:
        """Calculate a value for the STP0 ... STP7 registers of the AD9910 DDS.

        Parameters:
            ampl_scale:     Amplitude scale factor, unsigned 14-bit integer.
            phase_offset:   Phase offset, unsigned 16-bit integer, in steps of (2*pi / 2**16) radians.
            ftw:            Frequency tuning word, unsigned 32-bit integer, in steps of (1 GHz / 2**32).
        """
        if (not isinstance(ampl_scale, int)) or (ampl_scale < 0) or (ampl_scale >= 2**14):
            raise ValueError("Invalid ampl_scale parameter")
        if (not isinstance(phase_offset, int)) or (phase_offset < 0) or (phase_offset >= 2**16):
            raise ValueError("Invalid phase_offset parameter")
        if (not isinstance(ftw, int)) or (ftw < 0) or (ftw >= 2**32):
            raise ValueError("Invalid ftw parameter")
        return ((ampl_scale << 48)
                | (phase_offset << 32)
                | ftw)

    @staticmethod
    def _calc_ampl_scale(amplitude: float) -> int:
        """Convert amplitude scale factor from floating point range 0 ... 1.0 to 14-bit word for DDS."""
        return max(0, min(2**14 - 1, int(round(amplitude * (2**14 - 1)))))

    @staticmethod
    def _calc_phase_offset(phase: float) -> int:
        """Convert phase offset from floating point range 0 ... 1.0 to 16-bit word for DDS."""
        return max(0, min(2**16 - 1, int(round(phase * 2**16))))

    @staticmethod
    def _calc_ftw(frequency: float) -> int:
        """Convert frequency from floating point in Hz to 32-bit frequency tuning word."""
        return max(0, min(2**32 - 1, int(round(frequency / 1.0e9 * 2**32))))

    @staticmethod
    def _wrap_signed(value: int, wordlength: int) -> int:
        """Encode a signed integer as a two's complement representation in an unsigned word."""
        return value & (2**wordlength - 1)

    @rpc_method
    def dds_reset(self, channel: OutputChannel) -> None:
        """Reset both the DDS and the DCP for the specified channel(s).

        Resetting will disable the output channel and reset all DDS registers
        to their default settings. It will also stop the DCP and discard
        any pending DCP instructions.

        This command takes effect immediately.
        """
        self._transport.discard_read()  # Discard potential replies from previous commands.
        chan_str = self._format_channel(channel)
        self._write("dds {} reset".format(chan_str))

        # Wait for the response string from the reset command.
        while True:
            resp = self._transport.read_until(b"\n", self.COMMAND_RESPONSE_TIMEOUT)
            resp_str = resp.decode("latin1")
            if resp_str.startswith("DDS reset OK"):
                break

    @rpc_method
    def dcp_start(self, channel: OutputChannel) -> None:
        """Start the DCP for the specified channel(s).

        When started, the DCP will begin executing instructions from
        the DCP FIFO. The DCP will continue to run until it is explicitly
        disabled (via `dcp_stop()`) or until it is reset (via `dds_reset()`).

        Starting the DCP will also flush pending instructions from the
        input queue to the DCP FIFO.

        This command takes effect immediately.
        """
        chan_str = self._format_channel(channel)
        self._write("dcp {} start".format(chan_str))

    @rpc_method
    def dcp_stop(self, channel: OutputChannel) -> None:
        """Stop the DCP for the specified channel(s).

        This command takes effect immediately.
        """
        chan_str = self._format_channel(channel)
        self._write("dcp {} stop".format(chan_str))

    @rpc_method
    def dcp_spi_write(self,
                      channel: OutputChannel,
                      register: DdsRegister,
                      value: int,
                      wait_spi: bool = True,
                      flush: bool = False
                      ) -> None:
        """Send an SPI write instruction to the DCP.

        Instructs the DCP to write to the specified register of the AD9910 DDS.

        This instruction will be queued and only takes effect when it has
        been flushed to the DCP, and the DCP is running, and the DCP has
        completed all previous instructions.

        Parameters:
            channel:    Select one or both output channels.
            register:   Destination DDS register.
            value:      Value to write to the DDS register.
            wait_spi:   True to make the DCP wait until the SPI write is finished.
                        False to continue with the next DCP instruction while the
                        SPI write is pending.
            flush:      True to flush DCP instructions to the DCP FIFO.
        """
        chan_str = self._format_channel(channel)
        reg_str = DdsRegister(register).name
        if (not isinstance(value, int)) or (value >= 2**64):
            raise ValueError("Invalid value for register")
        wait_str = (":w" if wait_spi else ":c")
        flush_str = ("!" if flush else "")
        self._write("dcp {} spi:{}=0x{:x}{}{}".format(chan_str, reg_str, value, wait_str, flush_str))

    @rpc_method
    def dcp_register_write(self,
                           channel: OutputChannel,
                           register: DcpRegister,
                           value: int,
                           flush: bool = False
                           ) -> None:
        """Send a register write instruction to the DCP.

        Instructs the DCP to write to the specified DCP register.

        This instruction will be queued and only takes effect when it has
        been flushed to the DCP, and the DCP is running, and the DCP has
        completed all previous instructions.

        Parameters:
            channel:    Select one or both output channels.
            register:   Destination DCP register.
            value:      Value to write to the DCP register.
            flush:      True to flush DCP instructions to the DCP FIFO.
        """
        chan_str = self._format_channel(channel)
        reg_str = DcpRegister(register).name
        if (not isinstance(value, int)) or (value >= 2**32):
            raise ValueError("Invalid value for register")
        flush_str = ("!" if flush else "")
        self._write("dcp {} wr:{}=0x{:x}{}".format(chan_str, reg_str, value, flush_str))

    @rpc_method
    def dcp_update(self,
                   channel: OutputChannel,
                   flush: bool = False
                   ) -> None:
        """Send an update instruction to the DCP.

        Instructs the DCP to pulse the IO_UPDATE signal to the AD9910 DDS.
        A pulse on the IO_UPDATE signal will cause any new values in DDS
        registers to take effect.

        It is important that all SPI write instructions have completed
        before sending the IO_UPDATE pulse. This can be ensured by setting
        `wait_spi=True` on the last call to `dcp_spi_write()`.

        This instruction will be queued and only takes effect when it has
        been flushed to the DCP, and the DCP is running, and the DCP has
        completed all previous instructions.

        Parameters:
            channel:    Select one or both output channels.
            flush:      True to flush DCP instructions to the DCP FIFO.
        """
        chan_str = self._format_channel(channel)
        flush_str = ("!" if flush else "")
        self._write("dcp {} update:u{}".format(chan_str, flush_str))

    @rpc_method
    def set_single_tone(self,
                        channel: OutputChannel,
                        frequency: float,
                        amplitude: float,
                        phase: float
                        ) -> None:
        """Configure the specified channel to produce a single tone.

        The frequency will be rounded to the closest integer multiple of
        the internal frequency resolution (see `Wieserlabs_FlexDDS_NG_Dual.FREQUENCY_RESOLUTION`).

        The amplitude is a linear voltage scale factor, expressed as a fraction
        of the maximum output amplitude: +10 dBm (2 Vpp).

        This function assumes that the FlexDDS is in a "normal" configuration.
        When in doubt, call `dds_reset()` before calling this function.

        Parameters:
            channel:    Select one or both output channels.
            frequency:  Tone frequency in Hz (range 300 kHz ... 400 MHz).
            amplitude:  Amplitude scale factor (range 0.0 ... 1.0).
            phase:      Phase offset as fraction of the full sine wave period (range 0.0 ... 1.0).
        """

        if (frequency < 0) or (frequency > self.MAX_FREQUENCY):
            raise ValueError("Invalid frequency")
        if (amplitude < 0) or (amplitude > 1.0):
            raise ValueError("Invalid amplitude")
        if (phase < 0) or (phase > 1.0):
            raise ValueError("Invalid phase")

        cfr1 = self._calc_reg_cfr1()
        cfr2 = self._calc_reg_cfr2(ampl_scale_en=True,
                                   match_latency=True,
                                   hold_last_value=True,
                                   parallel_data_en=False)
        stp0 = self._calc_reg_stp(ampl_scale=self._calc_ampl_scale(amplitude),
                                  phase_offset=self._calc_phase_offset(phase),
                                  ftw=self._calc_ftw(frequency))

        self.dcp_spi_write(channel, DdsRegister.CFR1, cfr1)
        self.dcp_spi_write(channel, DdsRegister.CFR2, cfr2)
        self.dcp_spi_write(channel, DdsRegister.STP0, stp0, wait_spi=True)
        self.dcp_update(channel)
        self.dcp_start(channel)

    @rpc_method
    def set_amplitude_modulation(self,
                                 channel: OutputChannel,
                                 frequency: float,
                                 base_ampl: float,
                                 phase: float,
                                 mod_input: int,
                                 mod_offset: float,
                                 mod_scale: float
                                 ) -> None:
        """Configure the specified channel for a single tone with analog amplitude modulation.

        The `base_ampl` parameter sets the nominal amplitude scale factor,
        expressed as a fraction of the maximum output amplitude: +10 dBm (2 Vpp).
        This nominal amplitude applies when the modulation level is zero.

        The effective amplitude scale factor can be calculated as follows::

            A = base_ampl + mod_scale * (M + mod_offset)

        where `M` is the signal level of the analog modulation input in Volt (range -0.5 ... +0.5);
              `A` is the effective amplitude scale factor as a fraction of the maximum output level (2 Vpp).

        The effective amplitude scale factor `A` will be clipped to the range 0.0 ... 1.0.

        For example, the following call will configure output channel 0 for 10 MHz output
        with amplitude modulation to span the amplitude range from 0 Vpp (at analog input level -0.5 Volt)
        to 2 Vpp (at analog input level +0.5 Volt)::

            instr.set_amplitude_modulation(channel=OutputChannel.OUT0,
                                           frequency=10.0e6,
                                           base_amplitude=0.5,
                                           phase=0,
                                           mod_input=0,
                                           mod_offset=0,
                                           mod_scale=1.0)

        This function assumes that the FlexDDS is in a "normal" configuration.
        When in doubt, call `dds_reset()` before calling this function.

        Parameters:
            channel:    Select one or both output channels.
            frequency:  Tone frequency in Hz (range 300 kHz ... 400 MHz).
            base_ampl:  Base amplitude scale factor (range -2.0 ... 2.0).
            phase:      Phase offset as fraction of the full sine wave period (range 0.0 ... 1.0).
            mod_input:  Analog input channel for amplitude modulation (0 or 1).
            mod_offset: Offset to add to the analog input signal in Volt (range -0.5 ... +0.5).
            mod_scale:  Scale factor from analog input signal to RF output amplitude,
                        where 1.0 means that the full range of the analog input corresponds
                        to the full range of the output amplitude (range -30.0 ... +30.0).
        """

        if (frequency < 0) or (frequency > self.MAX_FREQUENCY):
            raise ValueError("Invalid frequency")
        if (base_ampl < -2.0) or (base_ampl > 2.0):
            raise ValueError("Invalid base_ampl parameter")
        if (phase < 0) or (phase > 1.0):
            raise ValueError("Invalid phase")
        if mod_input not in (0, 1):
            raise ValueError("Invalid mod_input parameter")
        if abs(mod_offset) > self.MAX_ANALOG_INPUT_VOLT:
            raise ValueError("Invalid mod_offset parameter")
        if (mod_scale < -30.0) or (mod_scale > 30.0):
            raise ValueError("Invalid mod_scale parameter")

        # Amplitude modulation requires the DCP and DDS working together as follows:
        # The DCP samples the analog input signal and converts the level to
        # an unsigned 16-bit word, based on scaling parameters that we specify.
        # The DCP feeds this 16-bit word to the DDS as "parallel data".
        # The DDS uses the 14 most significant bits of the parallel data word
        # as the effective amplitude scale factor.

        # Calculate DDS register values.
        # Enable "parallel data" to allow modulation.
        cfr1 = self._calc_reg_cfr1()
        cfr2 = self._calc_reg_cfr2(ampl_scale_en=True,
                                   match_latency=True,
                                   hold_last_value=True,
                                   parallel_data_en=True)
        stp0 = self._calc_reg_stp(ampl_scale=0,
                                  phase_offset=self._calc_phase_offset(phase),
                                  ftw=self._calc_ftw(frequency))

        # The ADC values from the modulation input are in range (-2**15 ... 2**15-1).
        # Scale the input offset to match that range.
        mod_input_offset_raw = int(round(mod_offset / self.MAX_ANALOG_INPUT_VOLT * 2**15))
        mod_input_offset_raw = self._wrap_signed(mod_input_offset_raw, 18)

        # The offset-corrected ADC values are multiplied by the raw scale factor,
        # then divided by 2**12. Calculate the raw scale factor to match
        # the specified effective scale factor.
        mod_scale_raw = int(round(mod_scale * 2**12))
        mod_scale_raw = self._wrap_signed(mod_scale_raw, 18)

        # A global offset is added to the scaled signal, then the result is
        # clipped to an unsigned 16-bit integer to form the parallel data word
        # for the DDS. The DDS uses the 14 most significant bits of the
        # parallel data word as the effective amplitude scale factor.
        # Scale the base amplitude to 16-bit range to determine the global offset value.
        mod_global_offset_raw = int(round(base_ampl * (2**16 - 4)))
        mod_global_offset_raw = self._wrap_signed(mod_global_offset_raw, 24)

        # Configure DDS.
        self.dcp_spi_write(channel, DdsRegister.CFR1, cfr1)
        self.dcp_spi_write(channel, DdsRegister.CFR2, cfr2)
        self.dcp_spi_write(channel, DdsRegister.STP0, stp0, wait_spi=True)

        # Configure modulation registers.
        self.dcp_register_write(channel, DcpRegister.AM_O, mod_global_offset_raw)
        if mod_input == 0:
            # Use analog input channel 0.
            self.dcp_register_write(channel, DcpRegister.AM_S1, 0)
            self.dcp_register_write(channel, DcpRegister.AM_O0, mod_input_offset_raw)
            self.dcp_register_write(channel, DcpRegister.AM_S0, mod_scale_raw)
        else:
            # Use analog input channel 1.
            self.dcp_register_write(channel, DcpRegister.AM_S0, 0)
            self.dcp_register_write(channel, DcpRegister.AM_O1, mod_input_offset_raw)
            self.dcp_register_write(channel, DcpRegister.AM_S1, mod_scale_raw)

        # Analog modulation config register:
        #   bit 29   = update modulation settings
        #   bits 1:0 = 00 = amplitude modulation
        mod_cfg = (1 << 29) | 0b00
        self.dcp_register_write(channel, DcpRegister.AM_CFG, mod_cfg)

        self.dcp_update(channel)
        self.dcp_start(channel)

    @rpc_method
    def set_frequency_modulation(self,
                                 channel: OutputChannel,
                                 base_freq: float,
                                 amplitude: float,
                                 phase: float,
                                 mod_input: int,
                                 mod_offset: float,
                                 mod_scale: float,
                                 ) -> None:
        """Configure the specified channel for a single tone with analog frequency modulation.

        The `base_freq` parameter sets the nominal output frequency in Hz.
        This nominal frequency applies when the modulation level is zero.

        The effective frequency can be calculated as follows::

            F = base_freq + mod_scale * (M + mod_offset)

        where `M` is the signal level of the analog modulation input in Volt (range -0.5 ... +0.5);
              `F` is the effective frequency in Hz.

        For example, the following call configures output channel 0 for a base output
        frequency of 10 MHz with a 1 MHz modulation range, such that analog input
        -0.5 V results in 9.5 MHz and analog input +0.5 V results in 10.5 MHz::

            instr.set_frequency_modulation(channel=OutputChannel.OUT0,
                                           base_freq=10.0e6,
                                           amplitude=0.5,
                                           phase=0,
                                           mod_input=0,
                                           mod_offset=0,
                                           mod_scale=1.0e6)

        This function assumes that the FlexDDS is in a "normal" configuration.
        When in doubt, call `dds_reset()` before calling this function.

        Parameters:
            channel:    Select one or both output channels.
            base_freq:  Base frequency in Hz (range 300 kHz ... 400 MHz).
            amplitude:  Amplitude scale factor (range 0.0 ... 1.0).
            phase:      Phase offset as fraction of the full sine wave period (range 0.0 ... 1.0).
            mod_input:  Analog input channel for amplitude modulation (0 or 1).
            mod_offset: Offset to add to the analog input signal in Volt (range -0.5 ... +0.5).
            mod_scale:  Frequency shift in Hz corresponding to the peak-to-peak range of the analog input
                        (range -500 MHz ... +500 MHz).
                        This can also be interpreted as a scale factor from analog input to frequency in Hz/Volt.
        """

        if (base_freq < 0) or (base_freq > self.MAX_FREQUENCY):
            raise ValueError("Invalid frequency")
        if (amplitude < 0) or (amplitude > 1.0):
            raise ValueError("Invalid amplitude")
        if (phase < 0) or (phase > 1.0):
            raise ValueError("Invalid phase")
        if mod_input not in (0, 1):
            raise ValueError("Invalid mod_input parameter")
        if abs(mod_offset) > self.MAX_ANALOG_INPUT_VOLT:
            raise ValueError("Invalid mod_offset parameter")
        if (mod_scale < -500.0e6) or (mod_scale > 500.0e6):
            raise ValueError("Invalid mod_scale parameter")

        # Frequency modulation requires the DCP and DDS working together as follows:
        # The DCP samples the analog input signal and converts the level to
        # an unsigned 16-bit word based on scaling parameters that we specify.
        # The DCP feeds this 16-bit word to the DDS as "parallel data".
        # The DDS takes the 16-bit word and shifts left by a shift count that
        # we specify (fm_gain). The result is added to the base FTW register
        # to obtain the effective FTW.

        # Calculate the FM gain (shift factor) required to span the modulation frequency range.
        # Choose fm_gain such that the 16-bit range of the parallel data word is enough
        # (with some margin) to cover the required frequency range.
        # Perhaps even the highest possible value of fm_gain (15) is not enough;
        # in that case the frequency modulation will clip to the achievable range.
        fm_gain = 0
        while (fm_gain < 15) and (((2**16 - 2) << fm_gain) * self.FREQUENCY_RESOLUTION < abs(mod_scale)):
            fm_gain += 1

        # Calculate the frequency tuning word for the base frequency.
        base_ftw = self._calc_ftw(base_freq)

        # Reduce the value of base_ftw such that the base frequency corresponds
        # to the middle of the 16-bit parallel data range (0x8000).
        if base_ftw >= (0x8000 << fm_gain):
            base_ftw -= (0x8000 << fm_gain)
        else:
            # Matching the base frequency to parallel data value 0x8000 requires
            # a negative value for base_ftw, which is not acceptable.
            # Instead choose base_ftw such that the base frequency corresponds to
            # some integer value of the parallel data word as close as possible to 0x8000.
            base_ftw &= ((1 << fm_gain) - 1)

        # Calculate DDS register values.
        # Enable "parallel data" to allow modulation.
        cfr1 = self._calc_reg_cfr1()
        cfr2 = self._calc_reg_cfr2(ampl_scale_en=True,
                                   match_latency=True,
                                   hold_last_value=True,
                                   parallel_data_en=True,
                                   fm_gain=fm_gain)
        stp0 = self._calc_reg_stp(ampl_scale=self._calc_ampl_scale(amplitude),
                                  phase_offset=self._calc_phase_offset(phase),
                                  ftw=base_ftw)

        # The ADC values from the modulation input are in range (-2**15 ... 2**15-1).
        # Scale the input offset to match that range.
        mod_input_offset_raw = int(round(mod_offset / self.MAX_ANALOG_INPUT_VOLT * 2**15))
        mod_input_offset_raw = self._wrap_signed(mod_input_offset_raw, 18)

        # Calculate the effective frequency step corresponding to a single
        # increment of the 16-bit parallel data word.
        mod_step = 2**fm_gain * self.FREQUENCY_RESOLUTION

        # The offset-corrected ADC values are multiplied by the raw scale factor,
        # then divided by 2**12. Calculate the raw scale factor such that the
        # peak-to-peak range of the ADC (2**16 - 1) gets scaled to the desired
        # frequency range.
        mod_scale_raw = int(round(mod_scale / mod_step / (2**16 - 1) * 2**12))
        mod_scale_raw = self._wrap_signed(mod_scale_raw, 18)

        # A global offset is added to the scaled signal, then the result is
        # clipped to an unsigned 16-bit integer to form the parallel data word
        # for the DDS. The DDS shifts and adds this word to the base FTW.
        # Choose the global offset such that it combines with the base FTW
        # to an effective FTW that matches the specified base frequency.
        mod_global_offset_raw = (self._calc_ftw(base_freq) - base_ftw) // 2**fm_gain

        # Configure DDS.
        self.dcp_spi_write(channel, DdsRegister.CFR1, cfr1)
        self.dcp_spi_write(channel, DdsRegister.CFR2, cfr2)
        self.dcp_spi_write(channel, DdsRegister.STP0, stp0)
        self.dcp_spi_write(channel, DdsRegister.FTW, base_ftw, wait_spi=True)

        # Configure modulation registers.
        self.dcp_register_write(channel, DcpRegister.AM_O, mod_global_offset_raw)
        if mod_input == 0:
            # Use analog input channel 0.
            self.dcp_register_write(channel, DcpRegister.AM_S1, 0)
            self.dcp_register_write(channel, DcpRegister.AM_O0, mod_input_offset_raw)
            self.dcp_register_write(channel, DcpRegister.AM_S0, mod_scale_raw)
        else:
            # Use analog input channel 1.
            self.dcp_register_write(channel, DcpRegister.AM_S0, 0)
            self.dcp_register_write(channel, DcpRegister.AM_O1, mod_input_offset_raw)
            self.dcp_register_write(channel, DcpRegister.AM_S1, mod_scale_raw)

        # Analog modulation config register:
        #   bit 29   = update modulation settings
        #   bits 1:0 = 10 = frequency modulation
        mod_cfg = (1 << 29) | 0b10
        self.dcp_register_write(channel, DcpRegister.AM_CFG, mod_cfg)

        self.dcp_update(channel)
        self.dcp_start(channel)

    @rpc_method
    def set_digital_modulation(self,
                               channel: OutputChannel,
                               frequency: float,
                               amplitude: float,
                               phase: float,
                               mod_input: int,
                               mod_invert: bool
                               ) -> None:
        """Configure the specified channel to produce a single tone with digital on/off modulation.

        The frequency will be rounded to the closest integer multiple of
        the internal frequency resolution (see `Wieserlabs_FlexDDS_NG_Dual.FREQUENCY_RESOLUTION`).

        The amplitude is a linear voltage scale factor, expressed as a fraction
        of the maximum output amplitude: +10 dBm (2 Vpp).

        This function assumes that the FlexDDS is in a "normal" configuration.
        When in doubt, call `dds_reset()` before calling this function.

        Parameters:
            channel:    Select one or both output channels.
            frequency:  Tone frequency in Hz (range 300 kHz ... 400 MHz).
            amplitude:  Amplitude scale factor (range 0.0 ... 1.0).
            phase:      Phase offset as fraction of the full sine wave period (range 0.0 ... 1.0).
            mod_input:  Digital input channel for on/off modulation (0 = BNC port A, 1 = BNC port B, 2 = BNC port C).
            mod_invert: True to invert the modulation signal (high TTL signal on BNC input disables the RF output);
                        False to use the non-inverted modulation signal (high TTL signal on BNC enables the RF output).
        """

        if (frequency < 0) or (frequency > self.MAX_FREQUENCY):
            raise ValueError("Invalid frequency")
        if (amplitude < 0) or (amplitude > 1.0):
            raise ValueError("Invalid amplitude")
        if (phase < 0) or (phase > 1.0):
            raise ValueError("Invalid phase")
        if mod_input not in (0, 1, 2):
            raise ValueError("Invalid modulation input channel")

        # Calculate DDS register values.
        # Enable manual OSK mode to allow switching between amplitude 0 and configured amplitude.
        ampl_scale = self._calc_ampl_scale(amplitude)
        cfr1 = self._calc_reg_cfr1(osk_enable=True,
                                   manual_osk=True,
                                   auto_osk=False)
        cfr2 = self._calc_reg_cfr2(ampl_scale_en=True,
                                   match_latency=True,
                                   hold_last_value=True,
                                   parallel_data_en=False)
        stp0 = self._calc_reg_stp(ampl_scale=ampl_scale,
                                  phase_offset=self._calc_phase_offset(phase),
                                  ftw=self._calc_ftw(frequency))

        # Switch the selected BNC port to input mode.
        #   bits 31:30 = WSCT = 00 = write register
        #   bit 9 = DIR       =  0 = input
        #   bit 8 = INV       =  0 = non-inverted
        #   bits 6:0 = OUT_MUX = 0 = not used in input mode
        bnc_reg = [DcpRegister.CFG_BNC_A,
                   DcpRegister.CFG_BNC_B,
                   DcpRegister.CFG_BNC_C][mod_input]
        self.dcp_register_write(OutputChannel.OUT0, bnc_reg, 0)

        # Route the BNC port to the OSK pin of the AD9910.
        cfg_osk = (2 << 9)                          # OPMODE = 2 (route MUX output to OSK)
        cfg_osk |= (1 << 8) if mod_invert else 0    # INV
        cfg_osk |= 5 + 3 * mod_input                # MUX = BNC_IN_[ABC]_LEVEL
        self.dcp_register_write(channel, DcpRegister.CFG_OSK, cfg_osk)

        # Configure DDS.
        # Note in OSK mode the AD9910 will take amplitude from the ASF register, not from STP0.
        self.dcp_spi_write(channel, DdsRegister.CFR1, cfr1)
        self.dcp_spi_write(channel, DdsRegister.CFR2, cfr2)
        self.dcp_spi_write(channel, DdsRegister.STP0, stp0)
        self.dcp_spi_write(channel, DdsRegister.ASF, ampl_scale << 2, wait_spi=True)

        self.dcp_update(channel)
        self.dcp_start(channel)

    @rpc_method
    def set_phase_modulation(self,
                             channel: OutputChannel,
                             frequency: float,
                             amplitude: float,
                             mod_input: int,
                             mod_offset: float,
                             mod_scale: float
                             ) -> None:
        """Configure the specified channel for a single tone with analog phase modulation.

        The effective phase can be calculated as follows::

            P = mod_scale * (M + mod_offset)

        where `M` is the signal level of the analog modulation input in Volt (range -0.5 ... +0.5);
              `P` is the effective phase as fraction of the full sine wave period (range 0.0 ... 1.0).

        The effective phase `P` will be clipped to the range 0.0 ... 1.0.
        It is thus not possible to shift the phase through multiple periods of the sine wave.

        This function assumes that the FlexDDS is in a "normal" configuration.
        When in doubt, call `dds_reset()` before calling this function.

        Parameters:
            channel:    Select one or both output channels.
            frequency:  Tone frequency in Hz (range 300 kHz ... 400 MHz).
            amplitude:  Amplitude scale factor (range 0.0 ... 1.0).
            mod_input:  Analog input channel for amplitude modulation (0 or 1).
            mod_offset: Offset to add to the analog input signal in Volt (range -0.5 ... +0.5).
            mod_scale:  Scale factor from analog input signal to RF phase,
                        where 1.0 means that the full range of the analog input corresponds
                        to one full sine wave period (range -30.0 ... +30.0).
        """

        if (frequency < 0) or (frequency > self.MAX_FREQUENCY):
            raise ValueError("Invalid frequency")
        if (amplitude < -2.0) or (amplitude > 2.0):
            raise ValueError("Invalid amplitude parameter")
        if mod_input not in (0, 1):
            raise ValueError("Invalid mod_input parameter")
        if abs(mod_offset) > self.MAX_ANALOG_INPUT_VOLT:
            raise ValueError("Invalid mod_offset parameter")
        if (mod_scale < -30.0) or (mod_scale > 30.0):
            raise ValueError("Invalid mod_scale parameter")

        # Phase modulation requires the DCP and DDS working together as follows:
        # The DCP samples the analog input signal and converts the level to
        # an unsigned 16-bit word, based on scaling parameters that we specify.
        # The DCP feeds this 16-bit word to the DDS as "parallel data".
        # The DDS uses the the parallel data word as the effective phase,
        # mapping word values 0 .. 2**16 to phase range 0 .. 2*Pi.

        # Calculate DDS register values.
        # Enable "parallel data" to allow modulation.
        cfr1 = self._calc_reg_cfr1()
        cfr2 = self._calc_reg_cfr2(ampl_scale_en=True,
                                   match_latency=True,
                                   hold_last_value=True,
                                   parallel_data_en=True)
        stp0 = self._calc_reg_stp(ampl_scale=self._calc_ampl_scale(amplitude),
                                  phase_offset=0,
                                  ftw=self._calc_ftw(frequency))

        # The ADC values from the modulation input are in range (-2**15 ... 2**15-1).
        # Scale the input offset to match that range.
        mod_input_offset_raw = int(round(mod_offset / self.MAX_ANALOG_INPUT_VOLT * 2**15))
        mod_input_offset_raw = self._wrap_signed(mod_input_offset_raw, 18)

        # The offset-corrected ADC values are multiplied by the raw scale factor,
        # then divided by 2**12. Calculate the raw scale factor to match
        # the specified effective scale factor.
        mod_scale_raw = int(round(mod_scale * 2**12))
        mod_scale_raw = self._wrap_signed(mod_scale_raw, 18)

        # Configure DDS.
        self.dcp_spi_write(channel, DdsRegister.CFR1, cfr1)
        self.dcp_spi_write(channel, DdsRegister.CFR2, cfr2)
        self.dcp_spi_write(channel, DdsRegister.STP0, stp0, wait_spi=True)

        # Configure modulation registers.
        mod_global_offset_raw = 0
        self.dcp_register_write(channel, DcpRegister.AM_O, mod_global_offset_raw)
        if mod_input == 0:
            # Use analog input channel 0.
            self.dcp_register_write(channel, DcpRegister.AM_S1, 0)
            self.dcp_register_write(channel, DcpRegister.AM_O0, mod_input_offset_raw)
            self.dcp_register_write(channel, DcpRegister.AM_S0, mod_scale_raw)
        else:
            # Use analog input channel 1.
            self.dcp_register_write(channel, DcpRegister.AM_S0, 0)
            self.dcp_register_write(channel, DcpRegister.AM_O1, mod_input_offset_raw)
            self.dcp_register_write(channel, DcpRegister.AM_S1, mod_scale_raw)

        # Analog modulation config register:
        #   bit 29   = update modulation settings
        #   bits 1:0 = 01 = phase modulation
        mod_cfg = (1 << 29) | 0b01
        self.dcp_register_write(channel, DcpRegister.AM_CFG, mod_cfg)

        self.dcp_update(channel)
        self.dcp_start(channel)



"""
Instrument driver for the WL Photonics tunable narrowband wavelength filter (WLTF-N).
"""

import logging
from dataclasses import dataclass
from re import search
from time import sleep

from qmi.core.context import QMI_Context
from qmi.core.exceptions import QMI_InstrumentException
from qmi.core.instrument import QMI_Instrument, QMI_InstrumentIdentification
from qmi.core.rpc import rpc_method
from qmi.core.transport import create_transport

# Global variable holding the logger for this module.
_logger = logging.getLogger(__name__)


@dataclass
class _WavelengthRange:
    """Dataclass for wavelength instrument range."""

    min: float
    max: float


@dataclass
class _StepsRange:
    """Dataclass for frequency instrument range."""

    min: int
    max: int


class WlPhotonics_WltfN(QMI_Instrument):
    """Instrument driver for the WL Photonics tunable narrowband wavelength filter.

    The instrument moves at limited speed.
    At power-up the instrument might start from motor position being 0, meaning that at this initial position
    the motor will be outside the (calibrated) wavelength range and reading the wavelength simply returns
    'Wavelength:Unknown'. Similarly, going to zero position will move the motor outside the wavelength range.
    Note that the wavelength and step ranges are inverse: Min steps value is max wavelength and vice versa.
    
    The default baudrate is 115200 for serial connections.

    Attributes:
        DEFAULT_RESPONSE_TIMEOUT:   Timeout value for waiting responses.
        ZEROING_WAIT:               Time to wait for the motor to go to zero.
        CMD_TERMINATOR:             The terminator added for sending commands.
        RESPONSE_TERMINATOR:        The terminator expected for end of one response.
    """
    # Default response timeout in seconds.
    DEFAULT_RESPONSE_TIMEOUT = 5.0
    ZEROING_WAIT = 3.0  # In seconds
    
    # Command and response terminator characters
    CMD_TERMINATOR = "\r\n"
    RESPONSE_TERMINATOR = b"OK\r\n"
    ENCODING = "ascii"

    def __init__(self, context: QMI_Context, name: str, transport: str) -> None:
        """Initialize the instrument driver.

        Parameters:
            context:    The QMI context for running the driver in.
            name:       Name for this instrument instance.
            transport:  QMI transport descriptor to connect to the instrument.
        """
        super().__init__(context, name)
        self._transport = create_transport(transport)

        # Instrument ranges for values. Will be updated in `open()`
        self._wavelength_range = _WavelengthRange
        self._steps_range = _StepsRange

        # Current motor position step value. Initialize with 0, it will be updated in `open()`
        self._motor_position = 0

        # For convenience
        self.response_terminator = self.RESPONSE_TERMINATOR.decode(self.ENCODING)

    def _write(self, cmd: str) -> None:
        """Write a command to the instrument and check if any errors or alerts were raised.

        Parameters:
            cmd: The input command to send.
        """
        self._check_is_open()
        cmd = cmd + self.CMD_TERMINATOR
        self._transport.write(cmd.encode(self.ENCODING))

    def _read(self) -> str:
        """Read data from the instrument buffer.

        Returns:
            response: Decoded response string.
        """
        self._check_is_open()
        response = self._transport.read_until(self.RESPONSE_TERMINATOR, timeout=self.DEFAULT_RESPONSE_TIMEOUT)
        return response.decode(self.ENCODING)

    def _clear_ok(self, cmd: str) -> None:
        """Clear response 'OK' from the instrument buffer. If the returned value was something else than "OK",
        the clear method will raise an exception.

        Parameters:
            cmd: The command that was sent in.

        Raises:
            QMI_InstrumentException: If the instrument returned another response than "OK".
        """
        response = self._read()
        # Check that we got 'OK' as second item in response. The first is the echo of the command sent.
        if response.split(self.response_terminator)[1] == "":
            return

        raise QMI_InstrumentException(f"Command {cmd} returned {response} instead of OK")

    def _ask_int(self, cmd: str) -> int:
        """Send a query and return an integer number response.

        Parameters:
            cmd:    The query command string.

        Raises:
            ValueError: If the response string was not an integer value or an empty string.
            TypeError: If there was no match for the response string to find numbers, giving `None`.

        Returns:
            response: The query response as integer.
        """
        self._write(cmd)
        full_response = self._read().split("\n")  # First item should be the response, the second "OK".
        try:
            return int(search(r"[-\d.]+", full_response[0])[0])  # type: ignore
        except ValueError as exc:
            raise QMI_InstrumentException(f"Unexpected response to command {cmd!r}: {full_response!r}") from exc
        except TypeError as exc:
            raise QMI_InstrumentException(f"Unexpected response to command {cmd!r}: {full_response!r}") from exc

    def _ask_float(self, cmd: str) -> float:
        """Send a query and return a floating point response.

        Parameters:
            cmd:    The query command string.

        Raises:
            ValueError: If the response string was not a floating point value or an empty string.
            TypeError: If there was no match for the response string to find numbers, giving `None`.

        Returns:
            response: The query response as a floating point.
        """
        self._write(cmd)
        full_response = self._read().split("\n")  # First item should be the response, the second "OK".
        if "Unknown" in full_response[0]:
            # The wavelength has not been set or calibrated yet
            _logger.warning(f"Trying to set a value with command {cmd}, but the parameter is not yet set or calibrated")
            return 0.0
        try:
            return float(search(r"[\d.]+", full_response[0])[0])  # type: ignore
        except ValueError as exc:
            raise QMI_InstrumentException(f"Unexpected response to command {cmd!r}: {full_response!r}") from exc
        except TypeError as exc:
            raise QMI_InstrumentException(f"Unexpected response to command {cmd!r}: {full_response!r}") from exc

    @rpc_method
    def open(self) -> None:
        _logger.info("[%s] Opening connection to instrument", self._name)
        self._check_is_closed()
        self._transport.open()
        super().open()
        # Check the ID to update the ranges and log
        _logger.info("Found instrument %s", self.get_idn())
        # Check the current motor step position
        self.get_motor_position()

    @rpc_method
    def close(self) -> None:
        _logger.info("[%s] Closing connection to instrument", self._name)
        self._check_is_open()
        super().close()
        self._transport.close()

    @rpc_method
    def get_idn(self) -> QMI_InstrumentIdentification:
        """Read instrument info and return QMI_InstrumentIdentification instance. The instrument returns several
        lines of data, example from the documentation:
        'WL200: SN(201307374), MD(2018-11-23)\r\n',
        'WL Range: 1021.509~1072.505nm(Step: 4654~556)\r\n',
        'OK\r\n'.
        This is processed to create the QMI_InstrumentIdentification object and also to update the wavelength and
        step range class attributes:
        - self._wavelength_range: Updates the instrument wavelength range min and max.
        - self._steps_range: Updates the instrument step range min and max.

        Returns:
            QMI_InstrumentIdentification: Data with e.g. idn.vendor = WL Photonics, idn.model = WL200,
            idn.serial = 201307374, idn.version = 2018-11-23.
        """
        self._write("dev?")
        resp = self._read()
        words = resp.rstrip().split("\r\n")
        if len(words) != 3:
            raise QMI_InstrumentException(f"Unexpected response to dev?, got {resp!r}")

        model = words[0].split(":")[0]
        serial = search(r"SN\([\d]+\)", words[0])[0].strip("SN()")  # type: ignore
        version = search(r"MD\([ \d-]+\)", words[0])[0].strip("MD()")  # type: ignore
        # Update the wavelength and step ranges on-the-go
        wl_range = words[1].split(":")[1].split("nm")[0].strip().split("~")
        self._wavelength_range.min = float(wl_range[0])
        self._wavelength_range.max = float(wl_range[1])
        step_range = words[1].split(":")[2].rstrip(f"){self.response_terminator}").lstrip().split("~")
        self._steps_range.max = int(step_range[0])  # NOTE: The range is inverse to the wavelength:
        self._steps_range.min = int(step_range[1])  # Minimum step value is maximum wavelength and vice versa.

        return QMI_InstrumentIdentification(vendor="WL Photonics", model=model, serial=serial, version=version)

    @rpc_method
    def set_center_wavelength(self, wavelength: float) -> None:
        """Set the center wavelength in nanometers.

        Parameters:
            wavelength: The target wavelength in nanometers.

        Raises:
            QMI_InstrumentException: If the wavelength is not in the instrument range.
        """
        unit = "nm"
        decimals = 3
        if wavelength < self._wavelength_range.min or wavelength > self._wavelength_range.max:
            raise ValueError(
                f"Wavelength {wavelength}{unit} out of instrument range "
                f"({self._wavelength_range.min}{unit} - {self._wavelength_range.max}{unit})"
            )

        cmd = f"wl{wavelength:.{decimals}f}"
        self._write(cmd)
        self._clear_ok(cmd)

    @rpc_method
    def get_center_wavelength(self) -> float:
        """Get the center wavelength. Default unit is nanometers.

        Returns:
            wavelength: The instrument wavelength in nanometers.
        """
        wavelength = self._ask_float("wl?")
        return wavelength

    @rpc_method
    def get_minimum_wavelength(self) -> float:
        """Get the minimum wavelength.

        Returns:
            self._wavelength_range.min: The instrument wavelength minimum in nanometers.
        """
        return self._wavelength_range.min

    @rpc_method
    def get_maximum_wavelength(self) -> float:
        """Get the maximum wavelength.

        Returns:
            self._wavelength_range.max: The instrument wavelength maximum in nanometers.
        """
        return self._wavelength_range.max

    @rpc_method
    def reverse_motor(self, steps: int) -> None:
        """Reverse the motor by input steps. This can be used to fine-tune the center wavelength.

        Parameters:
            steps: Number of steps the motor should reverse.

        Raises:
            ValueError: If calculated new motor position is not in the correct range.
        """
        new_position = self._motor_position - steps
        if new_position < self._steps_range.min:
            raise ValueError(
                f"Input value {steps} moves motor beyond minimum range {self._wavelength_range.min}. "
                f"The new motor position would have been {new_position}."
            )

        cmd = f"sb{steps}"
        self._write(cmd)
        self._clear_ok(cmd)

    @rpc_method
    def forward_motor(self, steps: int) -> None:
        """Move the motor forward by input steps. This can be used to fine-tune the center wavelength.

        Parameters:
            steps: Number of steps the motor should move forward.

        Raises:
            ValueError: If calculated new motor position is not in the correct range.
        """
        new_position = self._motor_position + steps
        if new_position > self._steps_range.max:
            raise ValueError(
                f"Input value {steps} moves motor beyond minimum range {self._wavelength_range.max}. "
                f"The new motor position would have been {new_position}."
            )

        cmd = f"sf{steps}"
        self._write(cmd)
        self._clear_ok(cmd)

    @rpc_method
    def get_motor_position(self) -> int:
        """Get the current step motor position.

        Attributes:
            self._motor_position: Updates the current motor position step value.

        Returns:
            self._motor_position: The current motor position step value.
        """
        self._motor_position = self._ask_int("s?")
        return self._motor_position

    @rpc_method
    def get_minimum_steps(self) -> int:
        """Get the minimum steps.

        Returns:
            self._steps_range.min: The instrument steps minimum.
        """
        return self._steps_range.min

    @rpc_method
    def get_maximum_steps(self) -> int:
        """Get the maximum steps.

        Returns:
            self._steps_range.max: The instrument steps maximum.
        """
        return self._steps_range.max

    @rpc_method
    def go_to_zero(self):
        """Move the motor to 'zero' position. This can take up to 3 seconds.

        The motor moves to zero, where the wavelength value cannot be read anymore as the motor position
        is outside the instrument's wavelength range. Reading the wavelength here gives "wavelength:Unknown" as
        response. This probably can be used as a calibration step for checking drift on absolute wavelength.
        """
        cmd = "z"
        self._write(cmd)
        self._clear_ok(cmd)
        # Wait for zeroing and then update the latest position
        sleep(self.ZEROING_WAIT)
        self.get_motor_position()



"""
Instrument driver for the Newport 843-R optical power meter.

This driver depends on PyUSB and works only on Linux.
"""

import sys
import time
import logging
import math

import typing
from typing import List, Optional, Tuple

from qmi.core.context import QMI_Context
from qmi.core.exceptions import QMI_InstrumentException
from qmi.core.instrument import QMI_Instrument, QMI_InstrumentIdentification
from qmi.core.rpc import rpc_method

if sys.platform.startswith("linux") or typing.TYPE_CHECKING:
    # Newport USB access is only supported on Linux.
    from qmi.instruments.newport.newport_843r_libusb import Newport_843R_libusb


# Global variable holding the logger for this module.
_logger = logging.getLogger(__name__)


class Newport_843R(QMI_Instrument):
    """Instrument driver for the Newport 843-R optical power meter."""

    _rpc_constants = ["SENSOR_TYPES"]

    # Meaning of sensor type codes returned by the instrument.
    SENSOR_TYPES = {
        "BT": "BeamTrack",
        "CP": "Pyroelectric",
        "SI": "Photodiode",
        "TH": "Termopile",
        "XX": "No sensor connected"
    }

    # Time in seconds to wait before each write and read operation.
    # The power meter needs some time between write and read operations
    # and between commands. Without sufficient delay, the power meter
    # may occasionally crash.
    COMMAND_DELAY = 0.05

    @staticmethod
    def list_instruments() -> List[str]:
        """Return a list of serial numbers for attached Newport 843-R instruments."""
        return Newport_843R_libusb.list_instruments()

    def __init__(self,
                 context: QMI_Context,
                 name: str,
                 serial_number: str
                 ) -> None:
        """Initialize driver.

        :argument name: Name for this instrument instance.
        :argument serial_number: USB serial number.
        """
        super().__init__(context, name)
        self._serial_number = serial_number
        self._instr = None  # type: Optional[Newport_843R_libusb]
        self._last_query_time: float = 0.0

    @rpc_method
    def open(self) -> None:
        _logger.info("[%s] Opening connection instrument serial_number=%r", self._name, self._serial_number)
        self._check_is_closed()
        self._instr = Newport_843R_libusb(self._serial_number)
        self._last_query_time = time.monotonic()
        super().open()

    @rpc_method
    def close(self) -> None:
        _logger.info("[%s] Closing connection to instrument", self._name)
        self._check_is_open()
        assert self._instr is not None
        self._instr.close()
        self._instr = None
        super().close()

    def ask(self, cmd: str) -> str:
        """Send command to instrument and return response from instrument."""
        self._check_is_open()
        assert self._instr is not None
        if not cmd.startswith("$"):
            raise QMI_InstrumentException("Command must start with '$'")
        rawcmd = cmd.encode('ascii') + b"\r\n"

        # Make sure there is enough time between commands.
        delay = self._last_query_time + self.COMMAND_DELAY - time.monotonic()
        if delay > 0:
            time.sleep(delay)

        self._instr.send_command(rawcmd)

        # Make sure there is enough time between write and read operations.
        time.sleep(self.COMMAND_DELAY)

        rawresp = self._instr.read_response(timeout=5)
        self._last_query_time = time.monotonic()

        resp = rawresp.decode('ascii', errors='replace')
        resp = resp.rstrip()
        if resp.startswith("?"):
            raise QMI_InstrumentException(f"Command '{cmd}' failed with error '{resp[1:]}'")
        if not resp.startswith("*"):
            raise QMI_InstrumentException(f"Unexpected response to '{cmd}', got '{resp}'")
        return resp

    @rpc_method
    def reset(self) -> None:
        """Reset the instrument, returning (most) settings to their defaults."""
        _logger.info("[%s] Resetting instrument", self._name)
        self.ask("$RE")

    @rpc_method
    def get_idn(self) -> QMI_InstrumentIdentification:
        """Read instrument type and version and return QMI_InstrumentIdentification instance."""
        instr_info = self.ask("$II")
        version_info = self.ask("$VE")
        words = instr_info[1:].strip().split()
        if len(words) != 3:
            raise QMI_InstrumentException(f"Unexpected response to $II, got '{instr_info}'")
        if not version_info.startswith("*EF"):
            raise QMI_InstrumentException(f"Unexpected response to $VE, got '{version_info}'")
        fw_version = version_info[3:]
        return QMI_InstrumentIdentification(vendor="Newport",
                                            model=words[2],
                                            serial=words[1],
                                            version=fw_version)

    @rpc_method
    def get_sensor_info(self) -> Tuple[str, str, str, str]:
        """Get sensor type and return information.

        :return: Tuple (sensor_type, serial_number, sensor_name, capabilities).
        """
        resp = self.ask("$HI")
        words = resp[1:].strip().split()
        if len(words) != 4:
            raise QMI_InstrumentException(f"Unexpected response to $HI, got '{resp}'")
        return (words[0], words[1], words[2], words[3])

    @rpc_method
    def get_filter_mode(self) -> bool:
        """Return True if the power meter is configured to work with a filter."""
        resp = self.ask("$FQ")
        words = resp[1:].split()
        i = int(words[0])
        if i < 1 or i >= len(words):
            raise QMI_InstrumentException(f"Unexpected response to $FQ, got '{resp}'")
        v = words[i].upper()
        if v not in ("IN", "OUT"):
            # Some types of sensors do not support filters and will return "* 1 N/A".
            # In this case we just report the filter is not enabled.
            _logger.warning("Unexpected response to $FQ, got %r", resp)
            return False
        return v == "IN"

    @rpc_method
    def set_filter_mode(self, filter_in: bool) -> None:
        """Configure the power meter to work with or without filter.

        If the filter configuration of the power meter does not match
        the actual presence of an attenuation filter, the meter will
        return wildly incorrect power measurements.

        :param filter_in: True to configure the meter with filter;
            False to configure the meter without a filter.
        """
        v = 2 if filter_in else 1
        self.ask(f"$FQ {v}")

    @rpc_method
    def get_power(self) -> float:
        """Return result of next power measurement (Watt).

        If the power meter sends an over-range response,
        this function returns +inf.

        This function is part of a generic power meter API
        and must be supported by all power meter implementations.
        """
        resp = self.ask("$SP")
        if resp[1:].upper() == "OVER":
            return math.inf
        return float(resp[1:])

    @rpc_method
    def get_timestamped_power(self) -> Tuple[float, float]:
        """Return a tuple of (unix_timestamp, measured_power).

        The timestamp is determined on the computer and may deviate
        from the actual time of the measurement.
        """
        resp = self.ask("$SP")
        timestamp = time.time()
        if resp[1:].upper() == "OVER":
            power = math.inf
        else:
            power = float(resp[1:])
        return (timestamp, power)

    @rpc_method
    def do_zero(self) -> None:
        """Perform a zeroing of internal circuits.

        Zeroing may take approximately 30 seconds.
        This function waits until zeroing is finished before returning.
        """
        resp = self.ask("$ZE")
        while True:
            time.sleep(0.25)
            resp = self.ask("$ZQ")
            status = resp[1:].strip()
            if not status.startswith("ZEROING IN PROG"):
                break
        if not status.startswith("ZEROING COMPLETE"):
            raise QMI_InstrumentException(f"Zeroing failed with error '{status}'")

    @rpc_method
    def zero_save(self) -> None:
        """Save the result of the preceding zero operation to device memory."""
        self.ask("$ZS")

    @rpc_method
    def get_available_ranges(self) -> List[str]:
        """Return a list of available range settings.

        Each range is formatted as a number with unit, e.g. "3.00uW".
        """
        resp = self.ask("$AR")
        words = resp[1:].strip().split()
        p = 1
        while p < len(words) and words[p].upper() in ("AUTO", "DBM"):
            p += 1
        return words[p:]

    @rpc_method
    def get_range(self) -> int:
        """Return the currently configured measuring range.

        The value -1 is returned if the instrument is in auto-range mode.
        Otherwise, this function returns an index into the list of available
        range settings, where 0 refers to the first available range.
        """
        resp = self.ask("$RN")
        return int(resp[1:].strip())

    @rpc_method
    def get_range_in_use(self) -> int:
        """Return the currently used measuring range.

        This function returns an index into the list of available range
        setting, where 0 refers to the first available range.

        When the instrument is in auto-range mode, this function returns
        the index of the automatically selected range.

        This function is part of a generic power meter API
        and must be supported by all power meter implementations.
        Other implementations may return the range as a floating point
        value in Watt.
        """
        resp = self.ask("$GU")
        return int(resp[1:].strip())

    @rpc_method
    def set_range(self, range_index: int) -> None:
        """Set the measurement range.

        Specify the value -1 to set the instrument to auto-range mode.
        Otherwise, specify an index into the list of available range settings,
        where 0 refers to the first available range.
        """
        self.ask(f"$WN {int(range_index)}")

    @rpc_method
    def set_autorange(self, enable: bool) -> None:
        """Enable or disable auto range mode.

        When autoranging is switched from enabled to disabled,
        the power meter will hold its last auto-selected range.

        This function is part of a generic power meter API
        and must be supported by all power meter implementations.
        """
        if enable:
            self.set_range(-1)
        else:
            selected_range = self.get_range_in_use()
            self.set_range(selected_range)

    @rpc_method
    def get_wavelength_range(self) -> Tuple[int, int]:
        """Return the range of available wavelengths.

        Returns a tuple (min_wavelength, max_wavelength).
        Wavelengths are expressed in nanometers.

        This command assumes that a continous-wavelength sensor is used.
        """
        resp = self.ask("$AW")
        words = resp[1:].strip().split()
        if len(words) < 4:
            raise QMI_InstrumentException(f"Unexpected response to $AW, got '{resp}'")
        if words[0].upper() != "CONTINUOUS":
            raise QMI_InstrumentException("Only continuous-wavelength sensors supported")
        wlow = int(words[1])
        whigh = int(words[2])
        return (wlow, whigh)

    @rpc_method
    def get_wavelength(self) -> int:
        """Return the current configured wavelength (nm).

        This command assumes that a continous-wavelength sensor is used.
        """
        resp = self.ask("$AW")
        words = resp[1:].strip().split()
        if len(words) < 4:
            raise QMI_InstrumentException(f"Unexpected response to $AW, got '{resp}'")
        if words[0].upper() != "CONTINUOUS":
            raise QMI_InstrumentException("Only continuous-wavelength sensors supported")
        index = int(words[3])
        if index < 1 or 3 + index >= len(words):
            raise QMI_InstrumentException(f"Unexpected response to $AW, got '{resp}'")
        wlen = int(words[3 + index])
        return wlen

    @rpc_method
    def set_wavelength(self, wavelength: int) -> None:
        """Configure the wavelength in nm.

        This function is part of a generic power meter API
        and must be supported by all power meter implementations.
        """
        self.ask(f"$WL {int(wavelength)}")

    @rpc_method
    def get_available_averaging(self) -> List[str]:
        """Return a list of available averaging modes.

        The first available mode is "NONE", i.e. no averaging.
        All other modes are formatted as a duration with unit, e.g. "3sec".
        """
        resp = self.ask("$AQ")
        words = resp[1:].strip().split()
        return words[1:]

    @rpc_method
    def get_averaging_mode(self) -> int:
        """Return the current averaging mode.

        This function returns an index into the list of available averaging
        modes, where 1 refers to the first available mode (which is "NONE").
        """
        resp = self.ask("$AQ")
        words = resp[1:].strip().split()
        return int(words[0])

    @rpc_method
    def set_averaging_mode(self, averaging_mode: int) -> None:
        """Set the averaging mode.

        Mode 1 is "NONE", i.e. no averaging.
        Otherwise, the specified value is an index into the list of
        available averaging modes, where 1 refers to the first available mode.
        """
        self.ask(f"$AQ {int(averaging_mode)}")



import qmi

from qmi.core.pubsub import QMI_SignalReceiver


def get_signals(recv_plot_signal: QMI_SignalReceiver):

    if recv_plot_signal.has_signal_ready():
        data = recv_plot_signal.get_next_signal()
        print(data)


if __name__ == "__main__":
    qmi.start("oh_no", context_cfg={"oh_yeah": {"tcp_server_port": 12345}})
    # First we need to find the context. If there are multiple contexts, filter by name.
    ctx = qmi.context().discover_peer_contexts()[0]  # The found context must be the one sending the data
    # Connect to it.
    qmi.context().connect_to_peer("oh_yeah", peer_address=ctx[1])
    # Get the RPC object from the connected context.
    ddc = qmi.get_rpc_object("oh_yeah.whoa")
    qmi.show_rpc_objects()
    # Create a signal receiver and subscribe to the publisher.
    sig_recv = QMI_SignalReceiver()
    ddc.signal_publisher.subscribe(sig_recv)
    # Then get signals and see that it works.
    while True:
        try:
            get_signals(sig_recv)

        except KeyboardInterrupt:
            break

        except:
            break

    try:
        qmi.context().disconnect_from_peer("oh_yeah")

    finally:
        qmi.stop()
        exit()
